# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from re import compile as Re

import os
import time

import isctest


def active(blob):
    return len([x for x in blob.splitlines() if " expiry" in x])


# global start-time variable
# pylint: disable=global-statement
START = 0


def test_initial():
    m = isctest.query.create("a.bogus.example.", "A")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.servfail(res)

    m = isctest.query.create("badds.example.", "SOA")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.servfail(res)

    m = isctest.query.create("a.secure.example.", "A")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.adflag(res)


def test_nta_validate_except(servers):
    ns4 = servers["ns4"]
    response = ns4.rndc("secroots -")
    assert Re("^corp: permanent") in response.out

    # check insecure local domain works with validate-except
    m = isctest.query.create("www.corp", "NS")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.noadflag(res)


def test_nta_bogus_lifetimes(servers):
    ns4 = servers["ns4"]

    # no nta lifetime specified:
    response = ns4.rndc("nta -l '' foo", raise_on_exception=False)
    assert "'nta' failed: bad ttl" in response.err

    # bad nta lifetime:
    response = ns4.rndc("nta -l garbage foo", raise_on_exception=False)
    assert "'nta' failed: bad ttl" in response.err

    # excessive nta lifetime:
    response = ns4.rndc("nta -l 7d1h foo", raise_on_exception=False)
    assert "'nta' failed: out of range" in response.err


def test_nta_install(servers):
    global START

    ns4 = servers["ns4"]
    ns4.rndc("nta -f -l 20s bogus.example")
    ns4.rndc("nta badds.example")

    # NTAs should persist after reconfig
    ns4.reconfigure()

    response = ns4.rndc("nta -d")
    assert len(response.out.splitlines()) == 3

    ns4.rndc("nta secure.example")
    ns4.rndc("nta fakenode.secure.example")
    with ns4.watch_log_from_here() as watcher:
        ns4.rndc("reload")
        watcher.wait_for_line("all zones loaded")

    response = ns4.rndc("nta -d")
    assert len(response.out.splitlines()) == 5

    START = time.time()


def test_nta_behavior(servers):
    assert START, "test_nta_behavior must be run as part of the full NTA test"

    m = isctest.query.create("a.bogus.example.", "A")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.noadflag(res)

    m = isctest.query.create("badds.example.", "SOA")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.noadflag(res)

    m = isctest.query.create("a.secure.example.", "A")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.noadflag(res)

    m = isctest.query.create("a.fakenode.secure.example.", "A")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noadflag(res)

    ns4 = servers["ns4"]
    response = ns4.rndc("secroots -")
    assert Re("^bogus.example: expiry") in response.out
    assert Re("^badds.example: expiry") in response.out
    assert Re("^secure.example: expiry") in response.out
    assert Re("^fakenode.secure.example: expiry") in response.out

    # secure.example and badds.example used the default nta-duration
    # (configured as 12s in ns4/named1.conf), but the nta recheck interval
    # is configured to 9s, so at t=10 the NTAs for secure.example and
    # fakenode.secure.example should both be lifted, while badds.example
    # should still be going.
    delay = START + 10 - time.time()
    if delay > 0:
        time.sleep(delay)

    m = isctest.query.create("b.secure.example", "A")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.adflag(res)

    m = isctest.query.create("b.fakenode.secure.example", "A")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.nxdomain(res)
    isctest.check.adflag(res)

    m = isctest.query.create("badds.example.", "SOA")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.noadflag(res)

    # bogus.example was set to expire in 20s, so at t=13
    # it should still be NTA'd, but badds.example used the default
    # lifetime of 12s, so it should revert to SERVFAIL now.
    delay = START + 13 - time.time()
    if delay > 0:
        time.sleep(delay)

    response = ns4.rndc("nta -d")
    assert active(response.out) <= 2

    response = ns4.rndc("secroots -")
    assert Re("bogus.example: expiry") in response.out
    assert Re("badds.example: expiry") not in response.out

    m = isctest.query.create("b.bogus.example", "A")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)

    m = isctest.query.create("a.badds.example", "A")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.servfail(res)
    isctest.check.noadflag(res)

    m = isctest.query.create("c.secure.example", "A")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.adflag(res)

    # at t=21, all the NTAs should have expired.
    delay = START + 21 - time.time()
    if delay > 0:
        time.sleep(delay)

    response = ns4.rndc("nta -d")
    assert active(response.out) == 0

    m = isctest.query.create("d.secure.example", "A")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.adflag(res)

    m = isctest.query.create("c.bogus.example", "A")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.servfail(res)
    isctest.check.noadflag(res)


def test_nta_removals(servers):
    ns4 = servers["ns4"]
    ns4.rndc("nta badds.example")

    response = ns4.rndc("nta -d")
    assert Re("^badds.example/_default: expiry") in response.out

    m = isctest.query.create("a.badds.example", "A")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.noadflag(res)

    response = ns4.rndc("nta -remove badds.example")
    assert "Negative trust anchor removed: badds.example" in response.out

    response = ns4.rndc("nta -d")
    assert Re("^badds.example/_default: expiry") not in response.out

    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.servfail(res)
    isctest.check.noadflag(res)

    # remove non-existent NTA three times
    ns4.rndc("nta -r foo")
    ns4.rndc("nta -remove foo")
    response = ns4.rndc("nta -r foo")
    assert "not found" in response.out


def test_nta_restarts(servers):
    global START
    assert START, "test_nta_restarts must be run as part of the full NTA test"

    # test NTA persistence across restarts
    ns4 = servers["ns4"]
    response = ns4.rndc("nta -d")
    assert active(response.out) == 0

    START = time.time()
    ns4.rndc("nta -f -l 30s bogus.example")
    ns4.rndc("nta -f -l 10s badds.example")
    response = ns4.rndc("nta -d")
    assert active(response.out) == 2

    # stop the server
    ns4.stop()

    # wait 14s before restarting. badds.example's NTA (lifetime=10s) should
    # have expired, and bogus.example should still be running.
    delay = START + 14 - time.time()
    if delay > 0:
        time.sleep(delay)
    ns4.start(["--noclean", "--restart", "--port", os.environ["PORT"]])

    response = ns4.rndc("nta -d")
    assert active(response.out) == 1
    assert Re("^bogus.example/_default: expiry") in response.out

    m = isctest.query.create("a.badds.example", "A")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.servfail(res)

    m = isctest.query.create("a.bogus.example", "A")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.noadflag(res)

    ns4.rndc("nta -r bogus.example")


def test_nta_regular(servers):
    global START
    assert START, "test_nta_regular must be run as part of the full NTA test"

    # check "regular" attribute in NTA file
    ns4 = servers["ns4"]

    response = ns4.rndc("nta -d")
    assert active(response.out) == 0

    # secure.example validates with AD=1
    m = isctest.query.create("a.secure.example", "A")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.adflag(res)

    # stop the server, update _default.nta, restart
    ns4.stop()
    now = time.localtime()
    future = str(now.tm_year + 20) + "0101010000"
    with open("ns4/_default.nta", "w", encoding="utf-8") as f:
        f.write(f"secure.example. regular {future}")

    ns4.start(["--noclean", "--restart", "--port", os.environ["PORT"]])

    # NTA active; secure.example. should now return an AD=0 answer.
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.noadflag(res)

    # nta-recheck is configured as 9s, so at t=12 the NTA for
    # secure.example. should be lifted as it is not a "forced" NTA.
    START = time.mktime(now)
    delay = START + 12 - time.time()
    if delay > 0:
        time.sleep(delay)

    response = ns4.rndc("nta -d")
    assert active(response.out) == 0

    # NTA lifted; secure.example. flush the cache to trigger a new query,
    # and it should now return an AD=1 answer.
    ns4.rndc("flushtree secure.example")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.adflag(res)


def test_nta_forced(servers):
    global START
    assert START, "test_nta_regular must be run as part of the full NTA test"

    # check "forced" attribute in NTA file
    ns4 = servers["ns4"]

    # just to be certain, clean up any existing NTA first
    ns4.rndc("nta -r secure.example")

    response = ns4.rndc("nta -d")
    assert active(response.out) == 0

    # secure.example validates with AD=1
    m = isctest.query.create("a.secure.example", "A")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.adflag(res)

    # stop the server, update _default.nta, restart
    ns4.stop()
    now = time.localtime()
    future = str(now.tm_year + 20) + "0101010000"
    with open("ns4/_default.nta", "w", encoding="utf-8") as f:
        f.write(f"secure.example. forced {future}")

    ns4.start(["--noclean", "--restart", "--port", os.environ["PORT"]])

    # NTA active; secure.example. should now return an AD=0 answer
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.noadflag(res)

    # nta-recheck is configured as 9s. at t=12 the NTA for
    # secure.example. should NOT be lifted as it is "forced".
    START = time.mktime(now)
    delay = START + 12 - time.time()
    if delay > 0:
        time.sleep(delay)

    # NTA lifted; secure.example. should still return an AD=0 answer
    ns4.rndc("flushtree secure.example")
    res = isctest.query.tcp(m, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.noadflag(res)


def test_nta_clamping(servers):
    ns4 = servers["ns4"]

    # clean up any existing NTA
    ns4.rndc("nta -r secure.example")

    # stop the server, update _default.nta, restart
    ns4.stop()
    now = time.localtime()
    future = str(now.tm_year + 20) + "0101010000"
    with open("ns4/_default.nta", "w", encoding="utf-8") as f:
        f.write(f"secure.example. forced {future}")

    ns4.start(["--noclean", "--restart", "--port", os.environ["PORT"]])

    # check that NTA lifetime read from file is clamped to 1 week.
    response = ns4.rndc("nta -d")
    assert active(response.out) == 1

    nta = next((s for s in response.out.splitlines() if " expiry" in s), None)
    assert nta is not None

    nta = nta.split(" ")
    expiry = f"{nta[2]} {nta[3]}"
    then = time.mktime(time.strptime(expiry, "%d-%b-%Y %H:%M:%S.000"))
    nextweek = time.mktime(now) + (86400 * 7)

    # normally there's no more than a few seconds difference between the
    # clamped expiration date and the calculated date for next week,
    # but add a 3600 second fudge factor to allow for daylight savings
    # changes.
    assert abs(nextweek - then < 3610)

    # remove the NTA
    ns4.rndc("nta -r secure.example")


def test_nta_forward(servers):
    ns9 = servers["ns9"]

    m = isctest.query.create("badds.example", "SOA")
    res = isctest.query.tcp(m, "10.53.0.9")
    isctest.check.servfail(res)
    isctest.check.empty_answer(res)
    isctest.check.noadflag(res)

    # add NTA and expect resolution to succeed
    ns9.rndc("nta badds.example")
    res = isctest.query.tcp(m, "10.53.0.9")
    isctest.check.noerror(res)
    isctest.check.rr_count_eq(res.answer, 2)
    isctest.check.noadflag(res)

    # remove NTA and expect resolution to fail again
    ns9.rndc("nta -remove badds.example")
    res = isctest.query.tcp(m, "10.53.0.9")
    isctest.check.servfail(res)
    isctest.check.empty_answer(res)
    isctest.check.noadflag(res)



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

"""
GL#5818 Finding 1 regression support — AsyncDnsServer primary.

Serves a minimal zone "sigaxfr.nil." whose AXFR carries two SIG records
at the same owner with different covered types (A and MX) and different
TTLs (600 and 1200).  A buggy secondary running dns_diff_load() with
rdata_covers() that only recognises RRSIG will file both rdatas under
typepair (SIG, 0) with the first tuple's TTL; a fixed secondary keeps
them under (SIG, A) and (SIG, MX) with their distinct TTLs.
"""

from collections.abc import AsyncGenerator

import dns.name
import dns.rcode
import dns.rdata
import dns.rdataclass
import dns.rdatatype
import dns.rrset

from isctest.asyncserver import (
    AsyncDnsServer,
    DnsResponseSend,
    DomainHandler,
    QueryContext,
    ResponseAction,
)

ZONE = dns.name.from_text("sigaxfr.nil.")
NS_NAME = dns.name.from_text("ns.sigaxfr.nil.")
HOST = dns.name.from_text("host.sigaxfr.nil.")

SOA_TEXT = "ns.sigaxfr.nil. hostmaster.sigaxfr.nil. 1 3600 1200 604800 3600"


def _make_sig_rdata(covered_text):
    """Produce a legacy SIG (24) rdata via RRSIG (46) round-trip."""
    rrsig = dns.rdata.from_text(dns.rdataclass.IN, dns.rdatatype.RRSIG, covered_text)
    wire = rrsig.to_digestable()
    return dns.rdata.from_wire(dns.rdataclass.IN, dns.rdatatype.SIG, wire, 0, len(wire))


class SigAxfrServer(DomainHandler):
    """Serve SOA and AXFR for sigaxfr.nil.; other qtypes get NOERROR/NODATA."""

    domains = ["sigaxfr.nil."]

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[ResponseAction, None]:
        soa_rrset = dns.rrset.from_text(
            ZONE, 3600, dns.rdataclass.IN, dns.rdatatype.SOA, SOA_TEXT
        )

        if qctx.qtype == dns.rdatatype.SOA:
            resp = qctx.response
            resp.answer.append(soa_rrset)
            yield DnsResponseSend(resp)
            return

        if qctx.qtype != dns.rdatatype.AXFR:
            # Other types: empty NOERROR response.
            yield DnsResponseSend(qctx.response)
            return

        # AXFR: opening SOA, NS, NS's A, two SIG RRs at the same owner
        # with distinct covered types and TTLs, closing SOA.
        resp = qctx.response
        resp.answer.append(soa_rrset)

        ns_rrset = dns.rrset.from_text(
            ZONE, 3600, dns.rdataclass.IN, dns.rdatatype.NS, str(NS_NAME)
        )
        resp.answer.append(ns_rrset)

        a_rrset = dns.rrset.from_text(
            NS_NAME, 3600, dns.rdataclass.IN, dns.rdatatype.A, "10.53.0.11"
        )
        resp.answer.append(a_rrset)

        sig_a = _make_sig_rdata("A 6 2 600 20260331170000 20260318160000 21831 . 0000")
        sig_a_rrset = dns.rrset.RRset(HOST, dns.rdataclass.IN, dns.rdatatype.SIG)
        sig_a_rrset.add(sig_a, ttl=600)
        resp.answer.append(sig_a_rrset)

        sig_mx = _make_sig_rdata(
            "MX 6 2 1200 20260331170000 20260318160000 21831 . 0000"
        )
        sig_mx_rrset = dns.rrset.RRset(HOST, dns.rdataclass.IN, dns.rdatatype.SIG)
        sig_mx_rrset.add(sig_mx, ttl=1200)
        resp.answer.append(sig_mx_rrset)

        # Closing SOA terminates the AXFR.
        resp.answer.append(soa_rrset)

        yield DnsResponseSend(resp)


def main() -> None:
    server = AsyncDnsServer(default_aa=True, default_rcode=dns.rcode.NOERROR)
    server.install_response_handler(SigAxfrServer())
    server.run()


if __name__ == "__main__":
    main()



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from typing import NamedTuple

import os


class Algorithm(NamedTuple):
    name: str
    number: int
    bits: int

    @classmethod
    def default(cls):
        return cls(
            os.environ["DEFAULT_ALGORITHM"],
            int(os.environ["DEFAULT_ALGORITHM_NUMBER"]),
            int(os.environ["DEFAULT_BITS"]),
        )


RSASHA1 = Algorithm("RSASHA1", 5, 2048)
NSEC3RSASHA1 = Algorithm("NSEC3RSASHA1", 7, 2048)
RSASHA256 = Algorithm("RSASHA256", 8, 2048)
RSASHA512 = Algorithm("RSASHA512", 10, 2048)
ECDSAP256SHA256 = Algorithm("ECDSAP256SHA256", 13, 256)
ECDSAP384SHA384 = Algorithm("ECDSAP384SHA384", 14, 384)
ED25519 = Algorithm("ED25519", 15, 256)
ED448 = Algorithm("ED448", 16, 456)

ALL_ALGORITHMS = [
    RSASHA1,
    NSEC3RSASHA1,
    RSASHA256,
    RSASHA512,
    ECDSAP256SHA256,
    ECDSAP384SHA384,
    ED25519,
    ED448,
]

ALL_ALGORITHMS_BY_NUM = {alg.number: alg for alg in ALL_ALGORITHMS}



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

"""
Regression tests for GL#5818: legacy DNSSEC types on the dynamic-update path.

SIG (24) and NXT (30) are obsolete DNSSEC record types, superseded by RRSIG
and NSEC in RFC 3755.  Allowing a client to inject them via dynamic update
exposed two bugs in sequence:

  * dns__db_findrdataset() asserted `covers == 0 || type == RRSIG`, which
    crashed named when a SIG update reached the prescan foreach_rr() call.
  * diff.c rdata_covers() dropped the covered type for SIG rdatas, so the
    zone DB stored every SIG rdataset under typepair (SIG, 0) instead of
    (SIG, covered_type); a second SIG add with a different covers and
    different TTL then tripped DNS_DBADD_EXACTTTL in qpzone and came back
    as SERVFAIL.

The adopted defence is to outright refuse SIG and NXT updates at the front
door (ns/update.c), keeping KEY updates permitted for SIG(0) transaction
signatures.  These tests verify the refusal.  The reachability of the
diff.c:rdata_covers() bug via inbound zone transfer is covered separately
by the AXFR-based regression test in this file.
"""

import time

import dns.rcode
import dns.rdata
import dns.rdataclass
import dns.rdataset
import dns.rdatatype
import dns.update
import pytest

import isctest

pytestmark = pytest.mark.extra_artifacts(
    [
        "ans*/ans.run",
        "ns*/*.bk",
        "ns*/*.conf",
        "ns*/*.db",
        "ns*/*.db.jnl",
        "ns*/*.db.signed",
        "ns*/*.jnl",
        "ns*/*.key",
        "ns*/K*.key",
        "ns*/K*.private",
        "ns*/K*.state",
        "ns*/dsset-*.",
        "ns6/sigaxfr.bk",
        "verylarge",
    ]
)


def _make_sig_rdata(text):
    """Create a SIG rdata from text.

    dnspython has no native text parser for the legacy SIG type (24),
    but the wire format is identical to RRSIG (46).  Parse as RRSIG,
    then re-wrap as SIG via the wire representation.
    """
    rrsig = dns.rdata.from_text(dns.rdataclass.IN, dns.rdatatype.RRSIG, text)
    wire = rrsig.to_digestable()
    return dns.rdata.from_wire(dns.rdataclass.IN, dns.rdatatype.SIG, wire, 0, len(wire))


def _make_nxt_rdata():
    """Create a minimal NXT rdata.

    NXT wire format (RFC 2535) is: next-name + type-bitmap.  The exact
    content does not matter for the refusal test; we just need a
    syntactically valid NXT rdata.
    """
    # next-name = root (\x00), type bitmap covering type A only.
    wire = b"\x00\x00\x00\x00\x40"
    return dns.rdata.from_wire(dns.rdataclass.IN, dns.rdatatype.NXT, wire, 0, len(wire))


def test_tcp_self_sig_record(ns6):
    """SIG (type 24) updates must be refused at the front door.

    Prior to the fix in dns__db_findrdataset(), a SIG update here
    crashed named.  Prior to the fix in diff.c rdata_covers(), the
    record was silently misfiled under typepair (SIG, 0).  The
    adopted policy outright refuses SIG (obsolete; use RRSIG) so the
    buggy dynamic-update paths are no longer reachable.  A PTR add
    first ensures the node exists, which is the original
    crash-reproducing precondition.
    """
    ptr_update = dns.update.UpdateMessage("in-addr.arpa.")
    ptr_update.add("1.0.53.10.in-addr.arpa.", 600, "PTR", "localhost.")
    response = isctest.query.tcp(
        ptr_update, ns6.ip, port=ns6.ports.dns, source="10.53.0.1"
    )
    assert response.rcode() == dns.rcode.NOERROR

    sig = _make_sig_rdata("A 6 0 86400 20260331170000 20260318160000 21831 . 0000")
    rds = dns.rdataset.Rdataset(dns.rdataclass.IN, dns.rdatatype.SIG)
    rds.update_ttl(600)
    rds.add(sig)
    sig_update = dns.update.UpdateMessage("in-addr.arpa.")
    sig_update.add("1.0.0.127.in-addr.arpa.", rds)

    with ns6.watch_log_from_here() as watcher:
        response = isctest.query.tcp(
            sig_update, ns6.ip, port=ns6.ports.dns, source="10.53.0.1"
        )
        assert response.rcode() == dns.rcode.REFUSED

        watcher.wait_for_line(
            "updating zone 'in-addr.arpa/IN': update failed: SIG updates are not allowed (REFUSED)"
        )

    # Confirm nothing of type SIG was stored.
    msg = isctest.query.create("1.0.53.10.in-addr.arpa.", "SIG")
    res = isctest.query.tcp(msg, ns6.ip, port=ns6.ports.dns)
    stored = any(rrset.rdtype == dns.rdatatype.SIG for rrset in res.answer)
    assert not stored, "SIG record was stored despite REFUSED response"


def test_tcp_self_nxt_record(ns6):
    """NXT (type 30) updates must be refused at the front door.

    NXT is the legacy DNSSEC denial-of-existence type, obsolete since
    RFC 3755 replaced it with NSEC.  Accepting it via dynamic update
    would let an authorised updater inject records that the signing
    and cut-point logic has no provision for.
    """
    # A second owner under a source that also matches tcp-self.
    source = "10.53.0.2"
    owner = "2.0.53.10.in-addr.arpa."

    ptr_update = dns.update.UpdateMessage("in-addr.arpa.")
    ptr_update.add(owner, 600, "PTR", "localhost.")
    response = isctest.query.tcp(ptr_update, ns6.ip, port=ns6.ports.dns, source=source)
    assert response.rcode() == dns.rcode.NOERROR

    nxt = _make_nxt_rdata()
    rds = dns.rdataset.Rdataset(dns.rdataclass.IN, dns.rdatatype.NXT)
    rds.update_ttl(600)
    rds.add(nxt)
    nxt_update = dns.update.UpdateMessage("in-addr.arpa.")
    nxt_update.add(owner, rds)

    with ns6.watch_log_from_here() as watcher:
        response = isctest.query.tcp(
            nxt_update, ns6.ip, port=ns6.ports.dns, source="10.53.0.1"
        )
        assert response.rcode() == dns.rcode.REFUSED

        watcher.wait_for_line(
            "updating zone 'in-addr.arpa/IN': update failed: NXT updates are not allowed (REFUSED)"
        )


def test_sig_covers_preserved_via_axfr(ns6):
    """Regression test for GL#5818 Finding 1, reached via AXFR.

    ans11 serves an AXFR for sigaxfr.nil. containing two SIG rdatas at
    the same owner with different covered types (A, MX) and different
    TTLs (600, 1200).  ns6 pulls the zone via dns_diff_load(), which
    calls diff.c rdata_covers(); before the fix that helper returned 0
    for SIG, so both tuples were grouped and filed under typepair
    (SIG, 0) with the first TTL (600) — the MX-covering record's TTL
    (1200) was silently dropped.  With the fix the records land in
    distinct typepairs and both TTLs survive.

    rndc dumpdb is used to inspect the secondary's stored state
    directly; the wire-level response can merge same-(owner,type,class)
    RRs and mask the difference.
    """
    zone = "sigaxfr.nil"
    owner = f"host.{zone}."
    dump_path = ns6.directory / "named_dump.db"

    # ns6 may have tried to SOA-poll ans11 before it was listening; force
    # a fresh refresh attempt and wait for the transfer to complete.
    with ns6.watch_log_from_here() as watcher:
        ns6.rndc(f"refresh {zone}")
        watcher.wait_for_line(f"zone {zone}/IN: transferred serial 1")

    # Remove any stale dump and ask named for a fresh one.
    if dump_path.exists():
        dump_path.unlink()
    ns6.rndc("dumpdb -zones")

    # rndc dumpdb is asynchronous; wait for the file and for its
    # trailing "Dump complete" marker.
    deadline_marker = "; Dump complete"
    for _ in range(50):
        if dump_path.exists():
            text = dump_path.read_text()
            if deadline_marker in text:
                break
        time.sleep(0.1)
    else:
        raise AssertionError(f"{dump_path} never contained {deadline_marker!r}")

    # Collect every SIG line for the owner from the dump.  Format is:
    #   <owner>. <ttl> IN SIG <covered> <alg> <labels> ...
    sig_lines = []
    for line in text.splitlines():
        fields = line.split()
        if len(fields) < 6:
            continue
        if not fields[0].lower().startswith("host.sigaxfr.nil"):
            continue
        if fields[2] != "IN" or fields[3] != "SIG":
            continue
        sig_lines.append(fields)

    assert (
        len(sig_lines) == 2
    ), f"expected 2 SIG records at {owner}, got {len(sig_lines)}: {sig_lines}"

    ttl_by_covers = {fields[4]: int(fields[1]) for fields in sig_lines}
    assert ttl_by_covers == {"A": 600, "MX": 1200}, (
        f"SIG records lost their covers/TTL binding: {ttl_by_covers}.  With "
        "the Finding 1 bug both records are filed under typepair (SIG, 0) "
        "and share the first-seen TTL (600)."
    )



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from re import compile as Re
from re import escape
from socket import AF_INET, SOCK_DGRAM, socket

import isctest


def run_attack(ns, name1, type1, name2, type2):
    msg1 = isctest.query.create(name1, type1, cd=True)
    msg2 = isctest.query.create(name2, type2, cd=True)
    port = int(isctest.vars.ALL["PORT"])

    with socket(AF_INET, SOCK_DGRAM) as sock:
        # The order the requests go out doesn't matter. What is important is
        # that the first query starts recursion before the second query returns
        # the answer, and the second query returns the answer before the first
        # query returns the answer. (So, when the NOERROR/NODATA comes back from
        # the first query, the cache is queried and we get the positive response
        # cached from the second query attached to the fresp rdataset of the
        # response of the first query.)
        # That ordering is enforced by ans2, which holds back the negative
        # answer to the first query until it has answered the second one (see
        # ans2/ans.py); the resolver must not crash while reconciling them.
        sock.sendto(msg1.to_wire(), (ns.ip, port))
        sock.sendto(msg2.to_wire(), (ns.ip, port))

    # The second query comes back immediately, the resolver caches the DNAME.
    # The first query comes back shortly after, once ans2 has released the
    # negative answer, and should not crash the server.  Wait for the negative
    # SOA for this specific name (not just any foo.test. one) so the test cannot
    # pass on an unrelated record.
    soa = Re(rf"(?<![\w.]){escape(name1)}.*IN\s+SOA\s+ns\.test\.\s+op\.ns\.test\.")
    with ns.watch_log_from_start(timeout=15) as watcher:
        watcher.wait_for_sequence([soa])



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from datetime import timedelta
from re import compile as Re

import os

import dns.name
import dns.rcode
import dns.rdataclass
import dns.rdatatype
import dns.update
import pytest

import isctest

pytestmark = pytest.mark.extra_artifacts(
    [
        "*.axfr",
        "*.created",
        "cdnskey.ns*",
        "cds.ns*",
        "dig.out.*",
        "rndc.dnssec.status.out.*",
        "secondary.cdnskey.ns*",
        "secondary.cds.ns*",
        "unused.*",
        "verify.out.*",
        "ns*/K*",
        "ns*/db-*",
        "ns*/keygen.out.*",
        "ns*/*.jbk",
        "ns*/*.jnl",
        "ns*/*.zsk",
        "ns*/*.signed",
        "ns*/*.journal.out.*",
        "ns*/settime.out.*",
        "ns*/model2.secondary.db",
    ]
)

CONFIG = {
    "dnskey-ttl": timedelta(hours=1),
    "ds-ttl": timedelta(days=1),
    "max-zone-ttl": timedelta(days=1),
    "parent-propagation-delay": timedelta(hours=1),
    "publish-safety": timedelta(hours=1),
    "retire-safety": timedelta(hours=1),
    "signatures-refresh": timedelta(days=5),
    "signatures-validity": timedelta(days=14),
    "zone-propagation-delay": timedelta(minutes=5),
}
TTL = 3600


def dsfromkey(key):
    dsfromkey_command = [
        os.environ.get("DSFROMKEY"),
        "-T",
        str(TTL),
        "-a",
        "SHA-256",
        "-C",
        "-w",
        str(key.keyfile),
    ]
    cmd = isctest.run.cmd(dsfromkey_command)
    return cmd.out.split()


def check_dnssec(server, zone, keys, expected):
    ksks = [k for k in keys if k.is_ksk()]
    zsks = [k for k in keys if not k.is_ksk()]

    isctest.kasp.check_keys(zone, keys, expected)

    for kp in expected:
        kp.set_expected_keytimes(CONFIG)
        kp.set_expected_keytimes(CONFIG)
        start = kp.key.get_timing("Created")
        kp.timing["Published"] = start
        kp.timing["Active"] = start
        if kp.role != "zsk":
            kp.timing["PublishCDS"] = start

    isctest.kasp.check_dnssec_verify(server, zone)
    isctest.kasp.check_apex(server, zone, ksks, zsks)


def check_no_dnssec_in_journal(server, zone):
    journalprint = [
        os.environ.get("JOURNALPRINT"),
        f"{server.identifier}/{zone}.db.jnl",
    ]

    cmd = isctest.run.cmd(journalprint)
    assert (
        Re(r"^\s*(?:\S+\s+){4}(NSEC|NSEC3|NSEC3PARAM|RRSIG)") not in cmd.out
    ), "dnssec record found in journal"


def wait_for_serial(primary, server, zone):
    if primary.identifier == server.identifier:
        # No need to check if the transfer has been done.
        return

    def check_serial():
        response = isctest.query.tcp(
            query, primary.ip, primary.ports.dns, timeout=3, attempts=1
        )
        assert response.rcode() == dns.rcode.NOERROR
        soa = response.get_rrset(
            response.answer,
            dns.name.from_text(fqdn),
            dns.rdataclass.IN,
            dns.rdatatype.SOA,
        )
        serial1 = soa[0].serial

        response = isctest.query.tcp(
            query, server.ip, server.ports.dns, timeout=3, attempts=1
        )
        assert response.rcode() == dns.rcode.NOERROR
        soa = response.get_rrset(
            response.answer,
            dns.name.from_text(fqdn),
            dns.rdataclass.IN,
            dns.rdatatype.SOA,
        )
        serial2 = soa[0].serial

        return (
            f"zone {zone}/IN (signed): serial {serial2} (unsigned {serial1})"
            in server.log
        )

    fqdn = f"{zone}."
    query = isctest.query.create(fqdn, dns.rdatatype.SOA)

    isctest.run.retry_with_timeout(check_serial, timeout=10)


def check_add_zsk(server, zone, keys, expected, extra_keys, extra, primary=None):
    if primary is None:
        primary = server

    isctest.log.info("add dnskey record:")

    isctest.log.info(
        f"- zone {zone} {primary.identifier}: update zone with ZSK from other providers"
    )

    update_msg = dns.update.UpdateMessage(zone)
    for zsk in extra_keys:
        dnskey = str(zsk.dnskey).split()
        rdata = " ".join(dnskey[4:])
        update_msg.add(f"{zone}.", TTL, "DNSKEY", rdata)
    primary.nsupdate(update_msg)

    wait_for_serial(primary, server, zone)

    # Check the new DNSKEY RRset.
    isctest.log.info(
        f"- zone {zone} {server.identifier}: check DNSKEY RRset after update add"
    )
    check_dnssec(server, zone, keys + extra_keys, expected + extra)

    # Check the logs for find zone keys errors.
    isctest.log.info(
        f"- zone {zone} {server.identifier}: make sure we did not try to sign with the keys added with nsupdate"
    )
    assert f"dns_zone_findkeys: error reading ./K{zone}" not in server.log

    # Trigger keymgr.
    with server.watch_log_from_here() as watcher:
        server.rndc(f"loadkeys {zone}")
        watcher.wait_for_line(f"keymgr: {zone} done")

    # Check again.
    isctest.log.info(f"- zone {zone} {server.identifier}: check again after keymgr run")
    check_dnssec(server, zone, keys + extra_keys, expected + extra)
    assert f"dns_zone_findkeys: error reading ./K{zone}" not in server.log


def _check_remove_zsk_fail(
    server, zone, keys, expected, extra_keys, extra, primary=None
):
    if primary is None:
        primary = server

    isctest.log.info(
        f"- zone {zone} {primary.identifier}: try to remove own ZSK (should fail)"
    )

    zsks = [k for k in keys if not k.is_ksk()]
    dnskey = str(zsks[0].dnskey).split()
    rdata = " ".join(dnskey[4:])
    update_msg = dns.update.UpdateMessage(zone)
    update_msg.delete(f"{zone}.", "DNSKEY", rdata)
    with primary.watch_log_from_here() as watcher:
        primary.nsupdate(update_msg)
        watcher.wait_for_line(
            f"updating zone '{zone}/IN': attempt to delete in use DNSKEY ignored"
        )

    # Both ZSKs should still be published.
    isctest.log.info(
        f"- zone {zone} {server.identifier}: check DNSKEY RRset after ignored remove"
    )
    check_dnssec(server, zone, keys + extra_keys, expected + extra)

    # Trigger keymgr.
    with server.watch_log_from_here() as watcher:
        server.rndc(f"loadkeys {zone}")
        watcher.wait_for_line(f"keymgr: {zone} done")

    # Check again.
    isctest.log.info(f"- zone {zone} {server.identifier}: check again after keymgr run")
    check_dnssec(server, zone, keys + extra_keys, expected + extra)


def check_remove_zsk(
    server,
    zone,
    keys,
    expected,
    extra_keys,
    extra,
    primary=None,
    check_fail=False,
    update_any=False,
):
    isctest.log.info("remove dnskey record:")

    if primary is None:
        primary = server

    if check_fail:
        _check_remove_zsk_fail(
            server,
            zone,
            keys,
            expected,
            extra_keys,
            extra,
            primary=primary,
        )

    if update_any:
        # Remove ZSK with update ANY.
        isctest.log.info(
            f"- zone {zone} {primary.identifier}: remove DNSKEY RRset with update ANY (expect ours)"
        )
        update_msg = dns.update.UpdateMessage(zone)
        update_msg.delete(f"{zone}.", "DNSKEY")
        primary.nsupdate(update_msg)
    else:
        # Remove actual ZSK.
        update_msg = dns.update.UpdateMessage(zone)
        for zsk in extra_keys:
            dnskey = str(zsk.dnskey).split()
            rdata = " ".join(dnskey[4:])
            update_msg.delete(f"{zone}.", "DNSKEY", rdata)
        primary.nsupdate(update_msg)

    wait_for_serial(primary, server, zone)

    # We should have only the KSK and ZSK from server.
    isctest.log.info(
        f"- zone {zone} {server.identifier}: check DNSKEY RRset after update remove"
    )
    check_dnssec(server, zone, keys, expected)

    # Trigger keymgr.
    with server.watch_log_from_here() as watcher:
        server.rndc(f"loadkeys {zone}")
        watcher.wait_for_line(f"keymgr: {zone} done")

    # Check again.
    isctest.log.info(f"- zone {zone} {server.identifier}: check again after keymgr run")
    check_dnssec(server, zone, keys, expected)


def check_add_cdnskey(server, zone, keys, expected, extra_keys, extra, primary=None):
    if primary is None:
        primary = server

    isctest.log.info("add cdnskey record:")

    isctest.log.info(
        f"- zone {zone} {primary.identifier}: update zone with CDNSKEY from other providers"
    )

    # Update the server with the CDNSKEY record from the other providers.
    update_msg = dns.update.UpdateMessage(zone)
    for ksk in extra_keys:
        dnskey = str(ksk.dnskey).split()
        rdata = " ".join(dnskey[4:])
        update_msg.add(f"{zone}.", TTL, "CDNSKEY", rdata)
    primary.nsupdate(update_msg)

    wait_for_serial(primary, server, zone)

    # Now there should be two CDNSKEY records.
    isctest.log.info(
        f"- zone {zone} {server.identifier}: check CDNSKEY RRset after update add"
    )
    check_dnssec(server, zone, keys + extra_keys, expected + extra)

    # Trigger keymgr.
    with server.watch_log_from_here() as watcher:
        server.rndc(f"loadkeys {zone}")
        watcher.wait_for_line(f"keymgr: {zone} done")

    # Check again.
    isctest.log.info(f"- zone {zone} {server.identifier}: check again after keymgr run")
    check_dnssec(server, zone, keys + extra_keys, expected + extra)


def _check_remove_cdnskey_fail(
    server, zone, keys, expected, extra_keys, extra, primary=None
):
    if primary is None:
        primary = server

    isctest.log.info(
        f"- zone {zone} {primary.identifier}: try to remove own CDNSKEY (should fail)"
    )

    ksks = [k for k in keys if not k.is_ksk()]
    dnskey = str(ksks[0].dnskey).split()
    rdata = " ".join(dnskey[4:])
    update_msg = dns.update.UpdateMessage(zone)
    update_msg.delete(f"{zone}.", "CDNSKEY", rdata)
    with primary.watch_log_from_here() as watcher:
        primary.nsupdate(update_msg)
        watcher.wait_for_line(
            f"updating zone '{zone}/IN': attempt to delete in use CDNSKEY ignored"
        )

    # Both CDNSKEY records should still be published.
    isctest.log.info(
        f"- zone {zone} {server.identifier}: check CDNSKEY RRset after ignored remove"
    )
    check_dnssec(server, zone, keys + extra_keys, expected + extra)

    # Trigger keymgr.
    with server.watch_log_from_here() as watcher:
        server.rndc(f"loadkeys {zone}")
        watcher.wait_for_line(f"keymgr: {zone} done")

    # Check again.
    isctest.log.info(f"- zone {zone} {server.identifier}: check again after keymgr run")
    check_dnssec(server, zone, keys + extra_keys, expected + extra)


def check_remove_cdnskey(
    server,
    zone,
    keys,
    expected,
    extra_keys,
    extra,
    primary=None,
    check_fail=False,
    update_any=False,
):
    isctest.log.info("remove cdnskey record:")

    if primary is None:
        primary = server

    if check_fail:
        _check_remove_cdnskey_fail(
            server,
            zone,
            keys,
            expected,
            extra_keys,
            extra,
            primary=primary,
        )

    if update_any:
        # Remove CDNSKEY with update ANY.
        isctest.log.info(
            f"- zone {zone} {primary.identifier}: remove CDNSKEY RRset with update ANY (expect ours)"
        )
        update_msg = dns.update.UpdateMessage(zone)
        update_msg.delete(f"{zone}.", "CDNSKEY")
        primary.nsupdate(update_msg)
    else:
        # Remove actual CDNSKEY.
        isctest.log.info(
            f"- zone {zone} {primary.identifier}: remove CDNSKEY from other providers"
        )

        update_msg = dns.update.UpdateMessage(zone)
        for ksk in extra_keys:
            dnskey = str(ksk.dnskey).split()
            rdata = " ".join(dnskey[4:])
            update_msg.delete(f"{zone}.", "CDNSKEY", rdata)
        primary.nsupdate(update_msg)

    wait_for_serial(primary, server, zone)

    # Now there should be one CDNSKEY record again.
    isctest.log.info(
        f"- zone {zone} {server.identifier}: check CDNSKEY RRset after update remove"
    )
    check_dnssec(server, zone, keys, expected)

    # Trigger keymgr.
    with server.watch_log_from_here() as watcher:
        server.rndc(f"loadkeys {zone}")
        watcher.wait_for_line(f"keymgr: {zone} done")

    # Check again.
    isctest.log.info(f"- zone {zone} {server.identifier}: check again after keymgr run")
    check_dnssec(server, zone, keys, expected)


def check_add_cds(server, zone, keys, expected, extra_keys, extra, primary=None):
    isctest.log.info("add cds record:")

    if primary is None:
        primary = server

    isctest.log.info(
        f"- zone {zone} {primary.identifier}: update zone with CDS from other providers"
    )

    # Update the server with the CDS record from the other providers.
    update_msg = dns.update.UpdateMessage(zone)
    for ksk in extra_keys:
        ds = dsfromkey(ksk)
        rdata = " ".join(ds[4:])
        update_msg.add(f"{zone}.", TTL, "CDS", rdata)
    primary.nsupdate(update_msg)

    wait_for_serial(primary, server, zone)

    # Now there should be two CDS records.
    isctest.log.info(
        f"- zone {zone} {server.identifier}: check CDS RRset after update add"
    )
    check_dnssec(server, zone, keys + extra_keys, expected + extra)

    # Trigger keymgr.
    with server.watch_log_from_here() as watcher:
        server.rndc(f"loadkeys {zone}")
        watcher.wait_for_line(f"keymgr: {zone} done")

    # Check again.
    isctest.log.info(f"- zone {zone} {server.identifier}: check again after keymgr run")
    check_dnssec(server, zone, keys + extra_keys, expected + extra)


def _check_remove_cds_fail(
    server, zone, keys, expected, extra_keys, extra, primary=None
):
    if primary is None:
        primary = server

    isctest.log.info(
        f"- zone {zone} {primary.identifier}: try to remove own CDS (should fail)"
    )

    ksks = [k for k in keys if not k.is_ksk()]
    ds = dsfromkey(ksks[0])
    rdata = " ".join(ds[4:])
    update_msg = dns.update.UpdateMessage(zone)
    update_msg.delete(f"{zone}.", "CDS", rdata)
    with primary.watch_log_from_here() as watcher:
        primary.nsupdate(update_msg)
        watcher.wait_for_line(
            f"updating zone '{zone}/IN': attempt to delete in use CDS ignored"
        )

    # Both CDS records should still be published.
    isctest.log.info(
        f"- zone {zone} {server.identifier}: check CDS RRset after ignored remove"
    )
    check_dnssec(server, zone, keys + extra_keys, expected + extra)

    # Trigger keymgr.
    with server.watch_log_from_here() as watcher:
        server.rndc(f"loadkeys {zone}")
        watcher.wait_for_line(f"keymgr: {zone} done")

    # Check again.
    isctest.log.info(f"- zone {zone} {server.identifier}: check again after keymgr run")
    check_dnssec(server, zone, keys + extra_keys, expected + extra)


def check_remove_cds(
    server,
    zone,
    keys,
    expected,
    extra_keys,
    extra,
    primary=None,
    check_fail=False,
    update_any=False,
):
    isctest.log.info("remove cds record:")

    if primary is None:
        primary = server

    if check_fail:
        _check_remove_cds_fail(
            server, zone, keys, expected, extra_keys, extra, primary=primary
        )

    if update_any:
        # Remove CDS with update ANY.
        isctest.log.info(
            f"- zone {zone} {primary.identifier}: remove CDS RRset with update ANY (expect ours)"
        )
        update_msg = dns.update.UpdateMessage(zone)
        update_msg.delete(f"{zone}.", "CDS")
        primary.nsupdate(update_msg)
    else:
        # Remove actual CDS.
        isctest.log.info(
            f"- zone {zone} {primary.identifier}: remove CDS from other providers"
        )

        update_msg = dns.update.UpdateMessage(zone)
        for ksk in extra_keys:
            ds = dsfromkey(ksk)
            rdata = " ".join(ds[4:])
            update_msg.delete(f"{zone}.", "CDS", rdata)
        primary.nsupdate(update_msg)

    wait_for_serial(primary, server, zone)

    # Now there should be one CDS record again.
    isctest.log.info(
        f"- zone {zone} {server.identifier}: check CDS RRset after update remove"
    )
    check_dnssec(server, zone, keys, expected)

    # Trigger keymgr.
    with server.watch_log_from_here() as watcher:
        server.rndc(f"loadkeys {zone}")
        watcher.wait_for_line(f"keymgr: {zone} done")

    # Check again.
    isctest.log.info(f"- zone {zone} {server.identifier}: check again after keymgr run")
    check_dnssec(server, zone, keys, expected)


def test_multisigner(ns3, ns4, default_algorithm):
    zone = "model2.multisigner"
    keyprops = [
        f"ksk 0 {default_algorithm.number} {default_algorithm.bits} goal:omnipresent dnskey:omnipresent krrsig:omnipresent ds:omnipresent",
        f"zsk 0 {default_algorithm.number} {default_algorithm.bits} goal:omnipresent dnskey:omnipresent zrrsig:omnipresent",
    ]

    # First make sure the zone is properly signed.
    isctest.log.info(f"basic DNSSEC tests for {zone}")
    isctest.kasp.wait_keymgr_done(ns3, zone)
    isctest.kasp.wait_keymgr_done(ns4, zone)

    keys3 = isctest.kasp.keydir_to_keylist(zone, ns3.identifier)
    ksks3 = [k for k in keys3 if k.is_ksk()]
    zsks3 = [k for k in keys3 if not k.is_ksk()]
    expected3 = isctest.kasp.policy_to_properties(ttl=TTL, keys=keyprops)

    check_dnssec(ns3, zone, keys3, expected3)

    keys4 = isctest.kasp.keydir_to_keylist(zone, ns4.identifier)
    ksks4 = [k for k in keys4 if k.is_ksk()]
    zsks4 = [k for k in keys4 if not k.is_ksk()]
    expected4 = isctest.kasp.policy_to_properties(ttl=TTL, keys=keyprops)

    check_dnssec(ns4, zone, keys4, expected4)

    # Add DNSKEY to RRset.
    newprops = [f"zsk unlimited {default_algorithm.number} {default_algorithm.bits}"]
    extra = isctest.kasp.policy_to_properties(ttl=TTL, keys=newprops)
    extra[0].private = False
    extra[0].legacy = True

    check_add_zsk(ns3, zone, keys3, expected3, [zsks4[0]], extra)
    check_add_zsk(ns4, zone, keys4, expected4, [zsks3[0]], extra)
    check_no_dnssec_in_journal(ns4, zone)

    # Remove DNSKEY from RRset.
    check_remove_zsk(ns3, zone, keys3, expected3, [zsks4[0]], extra, check_fail=True)
    check_remove_zsk(ns4, zone, keys4, expected4, [zsks3[0]], extra, check_fail=True)
    check_no_dnssec_in_journal(ns4, zone)

    # Add CDNSKEY RRset.
    newprops = [f"ksk unlimited {default_algorithm.number} {default_algorithm.bits}"]
    extra = isctest.kasp.policy_to_properties(ttl=TTL, keys=newprops)
    extra[0].private = False
    extra[0].legacy = True

    check_add_cdnskey(ns3, zone, keys3, expected3, [ksks4[0]], extra)
    check_add_cdnskey(ns4, zone, keys4, expected4, [ksks3[0]], extra)
    check_no_dnssec_in_journal(ns4, zone)

    # Remove CDNSKEY RRset.
    check_remove_cdnskey(
        ns3, zone, keys3, expected3, [ksks4[0]], extra, check_fail=True
    )
    check_remove_cdnskey(
        ns4, zone, keys4, expected4, [ksks3[0]], extra, check_fail=True
    )
    check_no_dnssec_in_journal(ns4, zone)

    # Update CDS RRset.
    check_add_cds(ns3, zone, keys3, expected3, [ksks4[0]], extra)
    check_add_cds(ns4, zone, keys4, expected4, [ksks3[0]], extra)
    check_no_dnssec_in_journal(ns4, zone)

    # Remove CDS RRset.
    check_remove_cds(ns3, zone, keys3, expected3, [ksks4[0]], extra, check_fail=True)
    check_remove_cds(ns4, zone, keys4, expected4, [ksks3[0]], extra, check_fail=True)
    check_no_dnssec_in_journal(ns4, zone)


def test_multisigner_update_any(ns3, ns4, default_algorithm):
    zone = "model2.update-any"
    keyprops = [
        f"ksk 0 {default_algorithm.number} {default_algorithm.bits} goal:omnipresent dnskey:omnipresent krrsig:omnipresent ds:omnipresent",
        f"zsk 0 {default_algorithm.number} {default_algorithm.bits} goal:omnipresent dnskey:omnipresent zrrsig:omnipresent",
    ]

    # First make sure the zone is properly signed.
    isctest.log.info(f"basic DNSSEC tests for {zone}")
    isctest.kasp.wait_keymgr_done(ns3, zone)
    isctest.kasp.wait_keymgr_done(ns4, zone)

    keys3 = isctest.kasp.keydir_to_keylist(zone, ns3.identifier)
    ksks3 = [k for k in keys3 if k.is_ksk()]
    zsks3 = [k for k in keys3 if not k.is_ksk()]
    expected3 = isctest.kasp.policy_to_properties(ttl=TTL, keys=keyprops)

    check_dnssec(ns3, zone, keys3, expected3)

    keys4 = isctest.kasp.keydir_to_keylist(zone, ns4.identifier)
    ksks4 = [k for k in keys4 if k.is_ksk()]
    zsks4 = [k for k in keys4 if not k.is_ksk()]
    expected4 = isctest.kasp.policy_to_properties(ttl=TTL, keys=keyprops)

    check_dnssec(ns4, zone, keys4, expected4)

    # Add DNSKEY to RRset.
    newprops = [f"zsk unlimited {default_algorithm.number} {default_algorithm.bits}"]
    extra = isctest.kasp.policy_to_properties(ttl=TTL, keys=newprops)
    extra[0].private = False
    extra[0].legacy = True

    check_add_zsk(ns3, zone, keys3, expected3, [zsks4[0]], extra)
    check_add_zsk(ns4, zone, keys4, expected4, [zsks3[0]], extra)
    check_no_dnssec_in_journal(ns4, zone)

    # Remove DNSKEY from RRset.
    check_remove_zsk(ns3, zone, keys3, expected3, [zsks4[0]], extra, update_any=True)
    check_remove_zsk(ns4, zone, keys4, expected4, [zsks3[0]], extra, update_any=True)
    check_no_dnssec_in_journal(ns4, zone)

    # Add CDNSKEY RRset.
    newprops = [f"ksk unlimited {default_algorithm.number} {default_algorithm.bits}"]
    extra = isctest.kasp.policy_to_properties(ttl=TTL, keys=newprops)
    extra[0].private = False
    extra[0].legacy = True

    check_add_cdnskey(ns3, zone, keys3, expected3, [ksks4[0]], extra)
    check_add_cdnskey(ns4, zone, keys4, expected4, [ksks3[0]], extra)
    check_no_dnssec_in_journal(ns4, zone)

    # Remove CDNSKEY RRset.
    check_remove_cdnskey(
        ns3, zone, keys3, expected3, [ksks4[0]], extra, update_any=True
    )
    check_remove_cdnskey(
        ns4, zone, keys4, expected4, [ksks3[0]], extra, update_any=True
    )
    check_no_dnssec_in_journal(ns4, zone)

    # Update CDS RRset.
    check_add_cds(ns3, zone, keys3, expected3, [ksks4[0]], extra)
    check_add_cds(ns4, zone, keys4, expected4, [ksks3[0]], extra)
    check_no_dnssec_in_journal(ns4, zone)

    # Remove CDS RRset.
    check_remove_cds(ns3, zone, keys3, expected3, [ksks4[0]], extra, update_any=True)
    check_remove_cds(ns4, zone, keys4, expected4, [ksks3[0]], extra, update_any=True)
    check_no_dnssec_in_journal(ns4, zone)


def test_multisigner_secondary(ns3, ns4, ns5, default_algorithm):
    zone = "model2.secondary"
    keyprops = [
        f"ksk 0 {default_algorithm.number} {default_algorithm.bits} goal:omnipresent dnskey:omnipresent krrsig:omnipresent ds:omnipresent",
        f"zsk 0 {default_algorithm.number} {default_algorithm.bits} goal:omnipresent dnskey:omnipresent zrrsig:omnipresent",
    ]

    # First make sure the zone is properly signed.
    isctest.log.info(f"basic DNSSEC tests for {zone}")
    isctest.kasp.wait_keymgr_done(ns3, zone)
    isctest.kasp.wait_keymgr_done(ns4, zone)

    keys3 = isctest.kasp.keydir_to_keylist(zone, ns3.identifier)
    ksks3 = [k for k in keys3 if k.is_ksk()]
    zsks3 = [k for k in keys3 if not k.is_ksk()]
    expected3 = isctest.kasp.policy_to_properties(ttl=TTL, keys=keyprops)

    check_dnssec(ns3, zone, keys3, expected3)

    keys4 = isctest.kasp.keydir_to_keylist(zone, ns4.identifier)
    ksks4 = [k for k in keys4 if k.is_ksk()]
    zsks4 = [k for k in keys4 if not k.is_ksk()]
    expected4 = isctest.kasp.policy_to_properties(ttl=TTL, keys=keyprops)

    check_dnssec(ns4, zone, keys4, expected4)

    # Add DNSKEY to RRset.
    newprops = [f"zsk unlimited {default_algorithm.number} {default_algorithm.bits}"]
    extra = isctest.kasp.policy_to_properties(ttl=TTL, keys=newprops)
    extra[0].private = False
    extra[0].legacy = True

    check_add_zsk(ns3, zone, keys3, expected3, [zsks4[0]], extra, primary=ns5)
    check_add_zsk(ns4, zone, keys4, expected4, [zsks3[0]], extra, primary=ns5)
    check_no_dnssec_in_journal(ns3, zone)
    check_no_dnssec_in_journal(ns4, zone)

    # Remove DNSKEY from RRset.
    check_remove_zsk(ns3, zone, keys3, expected3, [zsks4[0]], extra, primary=ns5)
    check_remove_zsk(ns4, zone, keys4, expected4, [zsks3[0]], extra, primary=ns5)
    check_no_dnssec_in_journal(ns3, zone)
    check_no_dnssec_in_journal(ns4, zone)

    # Add CDNSKEY RRset.
    newprops = [f"ksk unlimited {default_algorithm.number} {default_algorithm.bits}"]
    extra = isctest.kasp.policy_to_properties(ttl=TTL, keys=newprops)
    extra[0].private = False
    extra[0].legacy = True

    check_add_cdnskey(ns3, zone, keys3, expected3, [ksks4[0]], extra, primary=ns5)
    check_add_cdnskey(ns4, zone, keys4, expected4, [ksks3[0]], extra, primary=ns5)
    check_no_dnssec_in_journal(ns3, zone)
    check_no_dnssec_in_journal(ns4, zone)

    # Remove CDNSKEY RRset.
    check_remove_cdnskey(ns3, zone, keys3, expected3, [ksks4[0]], extra, primary=ns5)
    check_remove_cdnskey(ns4, zone, keys4, expected4, [ksks3[0]], extra, primary=ns5)
    check_no_dnssec_in_journal(ns3, zone)
    check_no_dnssec_in_journal(ns4, zone)

    # Update CDS RRset.
    check_add_cds(ns3, zone, keys3, expected3, [ksks4[0]], extra, primary=ns5)
    check_add_cds(ns4, zone, keys4, expected4, [ksks3[0]], extra, primary=ns5)
    check_no_dnssec_in_journal(ns3, zone)
    check_no_dnssec_in_journal(ns4, zone)

    # Remove CDS RRset.
    check_remove_cds(ns3, zone, keys3, expected3, [ksks4[0]], extra, primary=ns5)
    check_remove_cds(ns4, zone, keys4, expected4, [ksks3[0]], extra, primary=ns5)
    check_no_dnssec_in_journal(ns3, zone)
    check_no_dnssec_in_journal(ns4, zone)



#!/usr/bin/python3

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0

from isctest.asyncserver import AsyncDnsServer
from nsec_synthesis.ans1 import common, f004, f023


def main() -> None:
    keys = common.load_keys()
    server = AsyncDnsServer(default_aa=True)
    server.install_response_handler(f004.F004Handler(keys))
    server.install_response_handler(f023.F023Handler(keys))
    server.run()


if __name__ == "__main__":
    main()



#!/usr/bin/python3

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0

from collections.abc import AsyncGenerator
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from pathlib import Path

import base64

from cryptography.hazmat.primitives import serialization
from dns.rdtypes.dnskeybase import Flag

import dns.dnssec
import dns.flags
import dns.message
import dns.name
import dns.rcode
import dns.rdata
import dns.rdataclass
import dns.rdatatype
import dns.rrset

from isctest.asyncserver import DnsResponseSend, DomainHandler, QueryContext

TTL = 300
ZONE = "f217.test."
PEM_PATH = Path(f"{ZONE}.pem")
CHILD = f"evil.{ZONE}"
ATTACK = f"www.{CHILD}"
NSEC_OWNER = f"00000000.{CHILD}"
NSEC_NEXT = f"zzz.{CHILD}"
FORGED_A = "192.0.2.217"


@dataclass(frozen=True)
class Key:
    zone: dns.name.Name
    private_key: object
    dnskey: dns.rdata.Rdata


def name(text: str) -> dns.name.Name:
    return dns.name.from_text(text)


def load_key() -> Key:
    private_key = serialization.load_pem_private_key(
        PEM_PATH.read_bytes(), password=None
    )

    dnskey = dns.dnssec.make_dnskey(
        private_key.public_key(),
        dns.dnssec.Algorithm.ECDSAP256SHA256,
        flags=Flag.ZONE | Flag.SEP,
    )

    return Key(name(ZONE), private_key, dnskey)


def rrset(owner: str, rdtype: dns.rdatatype.RdataType, *rdatas: str) -> dns.rrset.RRset:
    return dns.rrset.from_text(owner, TTL, dns.rdataclass.IN, rdtype, *rdatas)


def rrset_from_rdata(owner: str, rdata: dns.rdata.Rdata) -> dns.rrset.RRset:
    return dns.rrset.from_rdata(name(owner), TTL, rdata)


def add_signed(
    section: list[dns.rrset.RRset], covered: dns.rrset.RRset, signer: Key
) -> None:
    rrsig = dns.dnssec.sign(
        covered,
        signer.private_key,
        signer.zone,
        signer.dnskey,
        lifetime=86400,
        verify=True,
    )
    section.append(covered)
    section.append(dns.rrset.from_rdata(covered.name, covered.ttl, rrsig))


def soa_rrset(zone: str) -> dns.rrset.RRset:
    return rrset(
        zone,
        dns.rdatatype.SOA,
        f"ns.{ZONE} hostmaster.{ZONE} 1 7200 3600 1209600 300",
    )


def garbage_rrsig(
    owner: str, covered: dns.rdatatype.RdataType, labels: int, signer: str
) -> dns.rrset.RRset:
    now = datetime.now(timezone.utc)
    inception = (now - timedelta(hours=1)).strftime("%Y%m%d%H%M%S")
    expiration = (now + timedelta(days=1)).strftime("%Y%m%d%H%M%S")
    signature = base64.b64encode(bytes(64)).decode("ascii")
    text = (
        f"{dns.rdatatype.to_text(covered)} 13 {labels} {TTL} "
        f"{expiration} {inception} 12345 {signer} {signature}"
    )
    rdata = dns.rdata.from_text(dns.rdataclass.IN, dns.rdatatype.RRSIG, text)
    return dns.rrset.from_rdata(name(owner), TTL, rdata)


def add_ds_denial(response: dns.message.Message, key: Key) -> None:
    add_signed(response.authority, soa_rrset(ZONE), key)
    nsec = rrset(CHILD, dns.rdatatype.NSEC, f"ns.{ZONE} NS RRSIG NSEC")
    add_signed(response.authority, nsec, key)


def add_attack_answer(response: dns.message.Message) -> None:
    """
    Crafted authoritative response to <q>.evil.f217.hack./A

        ;; ANSWER
        <q>.evil.f217.hack.        300 IN A     192.0.2.217
        <q>.evil.f217.hack.        300 IN RRSIG A 13 1 300 <exp> <inc> 12345 evil.f217.hack. <base64 of 64×0x00>
                                                    ^^^ Labels = 1, qname has 4 labels, wildcard heuristic fires

        ;; AUTHORITY (single owner, three rdatasets in this wire order)
        00000000.evil.f217.hack.   300 IN NSEC  zzz.evil.f217.hack. A RRSIG NSEC
        00000000.evil.f217.hack.   300 IN RRSIG NSEC 13 4 300 <exp> <inc> 12345 evil.f217.hack. <base64 of 64×0x00>
        00000000.evil.f217.hack.   300 IN NSEC3 1 0 0 - VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV A RRSIG
    """
    # A + RRSIG
    response.answer.append(rrset(ATTACK, dns.rdatatype.A, FORGED_A))
    response.answer.append(garbage_rrsig(ATTACK, dns.rdatatype.A, 1, CHILD))
    # NSEC
    nsec = rrset(
        NSEC_OWNER,
        dns.rdatatype.NSEC,
        f"{NSEC_NEXT} A RRSIG NSEC",
    )
    response.authority.append(nsec)
    # RRSIG(NSEC)
    response.authority.append(
        garbage_rrsig(
            NSEC_OWNER,
            dns.rdatatype.NSEC,
            len(name(NSEC_OWNER).labels) - 1,
            CHILD,
        )
    )
    # NSEC3
    nsec3 = rrset(
        NSEC_OWNER,
        dns.rdatatype.NSEC3,
        "1 0 0 - VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV A RRSIG",
    )
    response.authority.append(nsec3)


class RuntimeCheckHandler(DomainHandler):
    """Serve attacker.rrsig-labels-signer. with crafted wildcard RRSIG."""

    domains = [ZONE]

    def __init__(self) -> None:
        super().__init__()
        self.key = load_key()
        self.zone = name(ZONE)
        self.child = name(CHILD)
        self.attack = name(ATTACK)

    def match(self, qctx: QueryContext) -> bool:
        return qctx.qname.is_subdomain(self.zone)

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[DnsResponseSend, None]:
        qctx.prepare_new_response(with_zone_data=False)
        qctx.response.flags |= dns.flags.AA
        qctx.response.set_rcode(dns.rcode.NOERROR)

        if qctx.qname == self.zone and qctx.qtype == dns.rdatatype.DNSKEY:
            add_signed(
                qctx.response.answer,
                rrset_from_rdata(ZONE, self.key.dnskey),
                self.key,
            )
        elif qctx.qname == self.zone and qctx.qtype == dns.rdatatype.SOA:
            add_signed(qctx.response.answer, soa_rrset(ZONE), self.key)
        elif qctx.qname == self.child and qctx.qtype == dns.rdatatype.DS:
            add_ds_denial(qctx.response, self.key)
        elif qctx.qname == self.child and qctx.qtype == dns.rdatatype.DNSKEY:
            qctx.response.authority.append(soa_rrset(CHILD))
        elif qctx.qname == self.attack and qctx.qtype == dns.rdatatype.A:
            add_attack_answer(qctx.response)
        else:
            add_signed(qctx.response.authority, soa_rrset(ZONE), self.key)

        yield DnsResponseSend(qctx.response, authoritative=True)



#!/usr/bin/python3

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0

from pathlib import Path

import json

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec

import dns.dnssec
import dns.flags
import dns.name
import dns.rdataclass
import dns.rdatatype
import pytest

import isctest
import isctest.mark

F023_ZONE = "f023.test."
EVIL = f"evil.{F023_ZONE}"
POISON = f"0.{EVIL}"
POISON_NEXT = f"zzz.{F023_ZONE}"
TARGET = f"target.{F023_ZONE}"
TARGET_A = "192.0.2.77"
AUTH = "10.53.0.1"
RESOLVER = "10.53.0.2"

pytestmark = [
    isctest.mark.with_ecdsa_deterministic,
    pytest.mark.extra_artifacts(
        [
            "ans*/ans.run",
            "ans*/keys.json",
        ]
    ),
]


def _make_key(zone: str):
    private_key = ec.generate_private_key(ec.SECP256R1())
    dnskey = dns.dnssec.make_dnskey(
        private_key.public_key(),
        algorithm="ECDSAP256SHA256",
        flags=257,
    )
    ds = dns.dnssec.make_ds(dns.name.from_text(zone), dnskey, "SHA256")
    private_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    ).decode("ascii")
    return {
        "private_pem": private_pem,
        "dnskey": dnskey.to_text(),
        "ds": ds.to_text(),
    }


def bootstrap():
    keys = {F023_ZONE: _make_key(F023_ZONE)}
    Path("ans1/keys.json").write_text(json.dumps(keys, indent=2), encoding="ascii")
    dnskey = "".join(keys[F023_ZONE]["dnskey"].split()[3:])
    return {"DNSKEY": dnskey}


def _query(server, qname, qtype):
    query = isctest.query.create(qname, qtype)
    return isctest.query.tcp(query, server)


def _rrset(response, section, owner, rdtype, covers=None):
    if covers is None:
        return response.get_rrset(
            section,
            dns.name.from_text(owner),
            dns.rdataclass.IN,
            rdtype,
        )
    return response.get_rrset(
        section,
        dns.name.from_text(owner),
        dns.rdataclass.IN,
        rdtype,
        covers=covers,
    )


def _has_a(response, owner, address):
    rrset = _rrset(response, response.answer, owner, dns.rdatatype.A)
    return rrset is not None and any(rdata.address == address for rdata in rrset)


def _check_malicious_nsec(response, section):
    nsec = _rrset(response, section, POISON, dns.rdatatype.NSEC)
    assert nsec is not None, response.to_text()
    assert nsec[0].next == dns.name.from_text(POISON_NEXT), response.to_text()

    rrsig = _rrset(
        response,
        section,
        POISON,
        dns.rdatatype.RRSIG,
        covers=dns.rdatatype.NSEC,
    )
    assert rrsig is not None, response.to_text()
    assert rrsig[0].signer == dns.name.from_text(F023_ZONE), response.to_text()
    assert rrsig[0].key_tag == 12345, response.to_text()


def _prime_parent_for_aggressive_nsec():
    soa = _query(RESOLVER, F023_ZONE, "SOA")
    isctest.check.noerror(soa)
    isctest.check.adflag(soa)

    nsec = _query(RESOLVER, F023_ZONE, "NSEC")
    isctest.check.noerror(nsec)
    isctest.check.adflag(nsec)


def test_direct_insecure_child_nsec_next_fixture():
    poison = _query(AUTH, POISON, "NSEC")
    isctest.check.noerror(poison)
    assert poison.flags & dns.flags.AA
    _check_malicious_nsec(poison, poison.answer)


def test_resolver_does_not_synth_from_insecure_child_nsec():
    _prime_parent_for_aggressive_nsec()

    poison = _query(RESOLVER, POISON, "NSEC")
    isctest.check.noerror(poison)
    isctest.check.noadflag(poison)
    _check_malicious_nsec(poison, poison.answer)

    response = _query(RESOLVER, TARGET, "A")
    isctest.check.noerror(response)
    assert _has_a(response, TARGET, TARGET_A), response.to_text()
    isctest.check.adflag(response)
    assert _rrset(response, response.authority, POISON, dns.rdatatype.NSEC) is None



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

"""Handler for the attacker.rrsig-labels-signer. zone.

Serves crafted DNSSEC responses that exercise the RRSIG Labels underflow
check: an A query for any name under attacker.rrsig-labels-signer. returns:

  ANSWER:
    <qname> A 192.0.2.1
    <qname> RRSIG A 13 1 ...  (Labels=1, signed over *.rrsig-labels-signer.)

  AUTHORITY:
    <hash-1>.attacker.rrsig-labels-signer. NSEC3 ...  (covers
        hash(attacker.rrsig-labels-signer.) — next-closer proof for BIND)

The RRSIG has Labels=1 because the canonical signing input is
*.rrsig-labels-signer. A 192.0.2.1.  The signer name is
attacker.rrsig-labels-signer. (3 non-root labels), so Labels(1) is less
than signer_labels(3) - 1 = 2 — the condition the fix adds to
dns_dnssec_verify() in lib/dns/dnssec.c.

Key material is written by bootstrap() in tests_rrsig_labels_signer.py to
attacker_rrsig_labels_signer.pem in this directory before any server starts.
"""

from collections.abc import AsyncGenerator
from pathlib import Path

import base64
import hashlib
import logging
import time

from cryptography.hazmat.primitives import serialization
from dns.rdtypes.dnskeybase import Flag

import dns.dnssec
import dns.name
import dns.rcode
import dns.rdata
import dns.rdataclass
import dns.rdatatype
import dns.rdtypes.ANY.NSEC3
import dns.rrset

from isctest.asyncserver import (
    DnsResponseSend,
    DomainHandler,
    QueryContext,
    ResponseAction,
)

ZONE_NAME = "attacker.rrsig-labels-signer."
PARENT_ZONE_NAME = "rrsig-labels-signer."
POISON_IP = "192.0.2.1"
TTL = 300
PEM_PATH = Path("attacker_rrsig_labels_signer.pem")


def _to_b32hex_lower(data: bytes) -> str:
    """Base32hex-encode bytes, lowercase, no padding (NSEC3 owner-name form)."""
    return base64.b32hexencode(data).decode().rstrip("=").lower()


def _nsec3_sha1(name: str) -> bytes:
    """SHA-1 NSEC3 hash of name (0 iterations, empty salt)."""
    wire = dns.name.from_text(name).canonicalize().to_wire()
    return hashlib.sha1(wire).digest()


def _inc_bytes(b: bytes) -> bytes:
    out = bytearray(b)
    for i in range(len(out) - 1, -1, -1):
        out[i] += 1
        if out[i] != 0:
            break
    return bytes(out)


def _dec_bytes(b: bytes) -> bytes:
    out = bytearray(b)
    for i in range(len(out) - 1, -1, -1):
        if out[i] > 0:
            out[i] -= 1
            break
        out[i] = 0xFF
    return bytes(out)


def _type_bitmap(*types: int) -> bytes:
    """Build NSEC / NSEC3 type bitmap bytes (window 0, types 0–255)."""
    size = (max(types) >> 3) + 1
    bm = bytearray(size)
    for t in types:
        bm[t >> 3] |= 1 << (7 - (t & 7))
    return bytes(bm)


class AttackerZoneHandler(DomainHandler):
    """Serve attacker.rrsig-labels-signer. with crafted wildcard RRSIG."""

    domains = [ZONE_NAME]

    def __init__(self) -> None:
        super().__init__()
        self._zone = dns.name.from_text(ZONE_NAME)

        priv = serialization.load_pem_private_key(PEM_PATH.read_bytes(), password=None)

        self._dnskey = dns.dnssec.make_dnskey(
            priv.public_key(),
            dns.dnssec.Algorithm.ECDSAP256SHA256,
            flags=Flag.ZONE | Flag.SEP,
        )
        logging.info("attacker DNSKEY keytag: %d", dns.dnssec.key_id(self._dnskey))

        now = int(time.time())
        inception = now - 3600
        expiration = now + 14 * 86400

        # DNSKEY RRset and self-signature
        self._dnskey_rrset = dns.rrset.RRset(
            self._zone, dns.rdataclass.IN, dns.rdatatype.DNSKEY
        )
        self._dnskey_rrset.update_ttl(TTL)
        self._dnskey_rrset.add(self._dnskey)
        self._dnskey_rrsig = dns.dnssec.sign(
            self._dnskey_rrset,
            priv,
            signer=self._zone,
            dnskey=self._dnskey,
            inception=inception,
            expiration=expiration,
            lifetime=None,
            deterministic=False,  # for OpenSSL<3.2.0 compat
        )

        # Crafted RRSIG for *.rrsig-labels-signer. A  →  Labels=1
        wild_rrset = dns.rrset.RRset(
            dns.name.from_text(f"*.{PARENT_ZONE_NAME}"),
            dns.rdataclass.IN,
            dns.rdatatype.A,
        )
        wild_rrset.update_ttl(TTL)
        wild_rrset.add(
            dns.rdata.from_text(dns.rdataclass.IN, dns.rdatatype.A, POISON_IP)
        )
        self._crafted_a_rrsig = dns.dnssec.sign(
            wild_rrset,
            priv,
            signer=self._zone,
            dnskey=self._dnskey,
            inception=inception,
            expiration=expiration,
            lifetime=None,
            deterministic=False,  # for OpenSSL<3.2.0 compat
        )
        logging.info(
            "crafted A RRSIG labels=%d (expected 1)", self._crafted_a_rrsig.labels
        )

        # NSEC3 covering hash(attacker.rrsig-labels-signer.) for the
        # NEEDNOQNAME proof (validator.c:findnsec3proofs).
        h = _nsec3_sha1(ZONE_NAME)
        h_owner = _dec_bytes(h)  # owner hash just below target
        h_next = _inc_bytes(h)  # next  hash just above target

        owner_name = dns.name.from_text(f"{_to_b32hex_lower(h_owner)}.{ZONE_NAME}")
        nsec3_rdata = dns.rdtypes.ANY.NSEC3.NSEC3(
            rdclass=dns.rdataclass.IN,
            rdtype=dns.rdatatype.NSEC3,
            algorithm=1,  # SHA-1
            flags=0,
            iterations=0,
            salt=b"",
            next=h_next,
            windows=[(0, _type_bitmap(dns.rdatatype.A, dns.rdatatype.RRSIG))],
        )
        self._nsec3_rrset = dns.rrset.RRset(
            owner_name, dns.rdataclass.IN, dns.rdatatype.NSEC3
        )
        self._nsec3_rrset.update_ttl(TTL)
        self._nsec3_rrset.add(nsec3_rdata)

        nsec3_rrsig = dns.dnssec.sign(
            self._nsec3_rrset,
            priv,
            signer=self._zone,
            dnskey=self._dnskey,
            inception=inception,
            expiration=expiration,
            lifetime=None,
            deterministic=False,  # for OpenSSL<3.2.0 compat
        )
        self._nsec3_rrsig_rrset = dns.rrset.RRset(
            owner_name,
            dns.rdataclass.IN,
            dns.rdatatype.RRSIG,
            dns.rdatatype.NSEC3,
        )
        self._nsec3_rrsig_rrset.update_ttl(TTL)
        self._nsec3_rrsig_rrset.add(nsec3_rrsig)

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[ResponseAction, None]:
        qtype = qctx.qtype
        qname = qctx.qname
        response = qctx.prepare_new_response(with_zone_data=False)

        if qtype == dns.rdatatype.DNSKEY and qname == self._zone:
            response.set_rcode(dns.rcode.NOERROR)
            dnskey_rrsig_rrset = dns.rrset.RRset(
                self._zone,
                dns.rdataclass.IN,
                dns.rdatatype.RRSIG,
                dns.rdatatype.DNSKEY,
            )
            dnskey_rrsig_rrset.update_ttl(TTL)
            dnskey_rrsig_rrset.add(self._dnskey_rrsig)
            response.answer.extend([self._dnskey_rrset, dnskey_rrsig_rrset])
            yield DnsResponseSend(response, authoritative=True)

        elif qtype == dns.rdatatype.A and qname.is_subdomain(self._zone):
            response.set_rcode(dns.rcode.NOERROR)
            a_rrset = dns.rrset.RRset(qname, dns.rdataclass.IN, dns.rdatatype.A)
            a_rrset.update_ttl(TTL)
            a_rrset.add(
                dns.rdata.from_text(dns.rdataclass.IN, dns.rdatatype.A, POISON_IP)
            )
            a_rrsig_rrset = dns.rrset.RRset(
                qname,
                dns.rdataclass.IN,
                dns.rdatatype.RRSIG,
                dns.rdatatype.A,
            )
            a_rrsig_rrset.update_ttl(TTL)
            a_rrsig_rrset.add(self._crafted_a_rrsig)
            response.answer.extend([a_rrset, a_rrsig_rrset])
            response.authority.extend([self._nsec3_rrset, self._nsec3_rrsig_rrset])
            yield DnsResponseSend(response, authoritative=True)

        else:
            # REFUSED for anything we don't handle
            yield DnsResponseSend(response)



#!/usr/bin/python3

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0

from collections.abc import AsyncGenerator

import dns.flags
import dns.rcode
import dns.rdatatype
import dns.rrset

from dnssec_wildcard.ans1.common import (
    Key,
    add_dnskey,
    add_signed,
    name,
    rrset,
    soa_rrset,
    wildcard_rrsig,
)
from isctest.asyncserver import DnsResponseSend, DomainHandler, QueryContext

F043_ZONE = "f043.test."
F043_QUERY = f"svc.{F043_ZONE}"
F043_VICTIM = f"victim.{F043_ZONE}"

FORGED_A = "198.51.100.45"
LEGIT_A = "192.0.2.113"


def add_wildcard_a(section: list[dns.rrset.RRset], owner: str, key: Key) -> None:
    section.append(rrset(owner, dns.rdatatype.A, FORGED_A))
    section.append(wildcard_rrsig(owner, FORGED_A, key))


class F043Handler(DomainHandler):  # Additional from wildcard
    domains = [F043_ZONE]

    def __init__(self, keys: dict[str, Key]) -> None:
        super().__init__()
        self.keys = keys

        if F043_ZONE not in keys:
            return

        self.key = keys[F043_ZONE]
        self.zone = name(F043_ZONE)
        self.query = name(F043_QUERY)
        self.victim = name(F043_VICTIM)

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[DnsResponseSend, None]:
        qctx.prepare_new_response(with_zone_data=False)
        qctx.response.flags |= dns.flags.AA
        qctx.response.set_rcode(dns.rcode.NOERROR)

        if qctx.qname == self.zone and qctx.qtype == dns.rdatatype.DNSKEY:
            add_dnskey(qctx.response, self.zone, self.key)
        elif qctx.qname == self.zone and qctx.qtype == dns.rdatatype.SOA:
            add_signed(qctx.response.answer, soa_rrset(self.zone), self.key)
        elif qctx.qname == self.query and qctx.qtype == dns.rdatatype.MX:
            add_signed(
                qctx.response.answer,
                rrset(F043_QUERY, dns.rdatatype.MX, f"10 {F043_VICTIM}"),
                self.key,
            )
            add_wildcard_a(qctx.response.additional, F043_VICTIM, self.key)
        elif qctx.qname == self.victim and qctx.qtype == dns.rdatatype.A:
            add_signed(
                qctx.response.answer,
                rrset(F043_VICTIM, dns.rdatatype.A, LEGIT_A),
                self.key,
            )
        else:
            add_signed(qctx.response.authority, soa_rrset(self.zone), self.key)

        yield DnsResponseSend(qctx.response, authoritative=True)



"""
Copyright (C) Internet Systems Consortium, Inc. ("ISC")

SPDX-License-Identifier: MPL-2.0

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0.  If a copy of the MPL was not distributed with this
file, you can obtain one at https://mozilla.org/MPL/2.0/.

See the COPYRIGHT file distributed with this work for additional
information regarding copyright ownership.

For any query, returns a hand-crafted RRSIG whose Type-Covered field
is selected by the leftmost label of QNAME. The label is parsed as a
DNS type via `dns.rdatatype.from_text()`, so the resolver can be
probed with any meta-type by querying e.g. `any.attacker.test.`,
`axfr.attacker.test.`, `tsig.attacker.test.`, etc.
"""

from collections.abc import AsyncGenerator

import dns.flags
import dns.rcode
import dns.rdataclass
import dns.rdatatype
import dns.rrset

from isctest.asyncserver import (
    AsyncDnsServer,
    DnsResponseSend,
    QueryContext,
    ResponseHandler,
)


class RrsigCoversHandler(ResponseHandler):
    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[DnsResponseSend, None]:
        covers_label = qctx.qname.labels[0].decode("ascii").upper()
        covers = dns.rdatatype.from_text(covers_label)
        rrset = dns.rrset.from_text(
            qctx.qname,
            3600,
            dns.rdataclass.IN,
            dns.rdatatype.RRSIG,
            f"TYPE{int(covers)} 8 2 3600 20300101000000 20200101000000 "
            "12345 attacker.test. AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        )
        qctx.response.set_rcode(dns.rcode.NOERROR)
        qctx.response.flags |= dns.flags.AA
        qctx.response.answer.append(rrset)
        yield DnsResponseSend(qctx.response)


def main() -> None:
    server = AsyncDnsServer()
    server.install_response_handler(RrsigCoversHandler())
    server.run()


if __name__ == "__main__":
    main()



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

"""
Test that the resolver rejects DS records for sibling zones in referrals.

A custom authoritative server (ans4) returns a referral for
child.sibling-ds that includes a DS record for sibling.sibling-ds.  The
resolver must detect that the DS owner does not match the delegation NS
name and treat the response as a form error.
"""

from re import compile as Re

from dnssec_py.common import DNSSEC_PY_MARK
from isctest.template import NS2, Nameserver, zones
from isctest.zone import Zone, configure_root

import isctest

pytestmark = DNSSEC_PY_MARK

ANS4 = Nameserver("ans4")


def bootstrap():
    # Child zone on ns2 — the test queries a.child.sibling-ds which
    # resolves to the default template A record (10.0.0.1).
    child = Zone("child.sibling-ds", NS2)
    child.configure()

    # Sibling zone on ns2 — exists so the sibling DS in the referral
    # refers to a real delegation.
    sibling = Zone("sibling.sibling-ds", NS2)
    sibling.configure()

    # Parent zone rendered into ans4/ (subdir=None puts the .db file
    # directly in the ans4 directory where AsyncDnsServer loads it).
    parent = Zone("sibling-ds", ANS4, subdir=None)
    parent.delegations = [child, sibling]
    parent.configure()

    # Root zone delegates sibling-ds. to ans4.
    root = configure_root([parent])

    return {
        "trust_anchors": root.trust_anchors(),
        "zones": zones([root, child, sibling]),
    }


def test_sibling_ds_rejected(ns9):
    """Resolver must reject a referral that contains DS for a sibling zone."""
    log_ds_mismatch = Re(r"DS doesn't match the delegation owner name")

    msg = isctest.query.create("a.child.sibling-ds.", "A")

    with ns9.watch_log_from_here() as watcher:
        res = isctest.query.tcp(msg, ns9.ip)
        watcher.wait_for_line(log_ds_mismatch)

    isctest.check.servfail(res)



#!/usr/bin/env python3

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from functools import reduce
from resource import RLIMIT_NOFILE, getrlimit, setrlimit

import os
import random
import socket
import subprocess
import sys
import time

MULTIDIG_INSTANCES = 10
CONNECT_TRIES = 5

random.seed()

# Ensure we have enough file desriptors to work
rlimit_nofile = getrlimit(RLIMIT_NOFILE)
if rlimit_nofile[0] < 1024:
    setrlimit(RLIMIT_NOFILE, (1024, rlimit_nofile[1]))


# Introduce some random delay
def jitter():
    time.sleep((500 + random.randint(0, 250)) / 1000000.0)


# A set of simple procedures to get the test's configuration options
def get_http_port(http_secure=False):
    http_port_env = None
    if http_secure:
        http_port_env = os.getenv("HTTPSPORT")
    else:
        http_port_env = os.getenv("HTTPPORT")
    if http_port_env:
        return int(http_port_env)
    return 443


def get_http_host():
    bind_host = os.getenv("BINDHOST")
    if bind_host:
        return bind_host
    return "localhost"


def get_dig_path():
    dig_path = os.getenv("DIG")
    if dig_path:
        return dig_path
    return "dig"


# A simple class which creates the given number of TCP connections to
# the given host in order to stress the BIND's quota facility
class TCPConnector:
    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.connections = []

    def connect_one(self):
        tries = CONNECT_TRIES
        while tries > 0:
            try:
                sock = socket.create_connection(
                    address=(self.host, self.port), timeout=None
                )
                self.connections.append(sock)
                break
            except ConnectionResetError:
                # some jitter for BSDs
                jitter()
                continue
            except TimeoutError:
                jitter()
                continue
            finally:
                tries -= 1

    # Close an established connection (randomly)
    def disconnect_random(self):
        pos = random.randint(0, len(self.connections) - 1)
        conn = self.connections[pos]
        try:
            conn.shutdown(socket.SHUT_RDWR)
            conn.close()
        except OSError:
            conn.close()
        finally:
            self.connections.remove(conn)

    def disconnect_all(self):
        while len(self.connections) != 0:
            self.disconnect_random()


# A simple class which allows running a dig instance under control of
# the process
class SubDIG:
    def __init__(self, http_secure=None, extra_args=None):
        self.sub_process = None
        self.dig_path = get_dig_path()
        self.host = get_http_host()
        self.port = get_http_port(http_secure=http_secure)
        if http_secure:
            self.http_secure = True
        else:
            self.http_secure = False
        self.extra_args = extra_args

    # This method constructs a command string
    def get_command(self):
        command = self.dig_path + " -p " + str(self.port) + " "
        command = command + "+noadd +nosea +nostat +noquest +nocmd +time=30 "
        if self.http_secure:
            command = command + "+https "
        else:
            command = command + "+http-plain "
        command = command + "@" + self.host + " "
        if self.extra_args:
            command = command + self.extra_args
        return command

    def run(self):
        with open(os.devnull, "w", encoding="utf-8") as devnull:
            self.sub_process = subprocess.Popen(  # pylint: disable=consider-using-with
                self.get_command(), shell=True, stdout=devnull
            )

    def wait(self, timeout=None):
        res = None
        if timeout is None:
            return self.sub_process.wait()
        try:
            res = self.sub_process.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            return None
        return res

    def alive(self):
        return self.sub_process.poll() is None


# A simple wrapper class which allows running multiple dig instances
# and examining their statuses in one logical operation.
class MultiDIG:
    def __init__(self, numdigs, http_secure=None, extra_args=None):
        assert int(numdigs) > 0, f"numdigs={numdigs}"
        digs = []
        for _ in range(1, int(numdigs) + 1):
            digs.append(SubDIG(http_secure=http_secure, extra_args=extra_args))
        self.digs = digs
        assert len(self.digs) == int(numdigs), f"len={len(self.digs)} numdigs={numdigs}"

    def run(self):
        for p in self.digs:
            p.run()

    def wait(self):
        return map(lambda p: (p.wait()), self.digs)

    # Wait for the all instances to terminate with expected given
    # status. Returns true or false.
    def wait_for_result(self, result):
        return reduce(
            lambda a, b: ((a == result or a is True) and b == result), self.wait()
        )

    def alive(self):
        return reduce(lambda a, b: (a and b), map(lambda p: (p.alive()), self.digs))

    def completed(self):
        total = 0
        for p in self.digs:
            if not p.alive():
                total += 1
        return total


# The test's main logic
def run_test(http_secure=True):
    query_args = "SOA ."
    # Let's try to make a successful query
    subdig = SubDIG(http_secure=http_secure, extra_args=query_args)
    subdig.run()
    assert subdig.wait() == 0, "DIG was expected to succeed"
    # Let's create a lot of TCP connections to the server stress the
    # HTTP quota
    connector = TCPConnector(get_http_host(), get_http_port(http_secure=http_secure))
    # Let's make queries until the quota kicks in
    subdig = SubDIG(http_secure=http_secure, extra_args=query_args)
    subdig.run()
    while True:
        connector.connect_one()
        subdig = SubDIG(http_secure=http_secure, extra_args=query_args)
        subdig.run()
        if subdig.wait(timeout=5) is None:
            break

    # At this point quota has kicked in.  Additionally, let's create a
    # bunch of dig processes all trying to make a query against the
    # server with exceeded quota
    multidig = MultiDIG(
        MULTIDIG_INSTANCES, http_secure=http_secure, extra_args=query_args
    )
    multidig.run()
    # Wait for the dig instance to complete. Not a single instance has
    # a chance to complete successfully because of the exceeded quota
    assert (
        subdig.wait(timeout=5) is None
    ), "The single DIG instance has stopped prematurely"
    assert subdig.alive(), "The single DIG instance is expected to be alive"
    assert multidig.alive(), (
        "The DIG instances from the set are all expected to "
        f"be alive, but {multidig.completed()} of them have completed"
    )
    # Let's close opened connections (in random order) to let all dig
    # processes to complete
    connector.disconnect_all()
    # Wait for all processes to complete successfully
    assert subdig.wait() == 0, "Single DIG instance failed"
    assert (
        multidig.wait_for_result(0) is True
    ), "One or more of DIG instances returned unexpected results"


def main():
    run_test(http_secure=True)
    run_test(http_secure=False)
    # If we have reached this point we could safely return 0
    # (success). If the test fails because of an assert, the whole
    # program will return non-zero exit code and produce the backtrace
    return 0


sys.exit(main())



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from pathlib import Path
from typing import TypeAlias

import base64
import shutil

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec, ed448, ed25519, rsa
from dns.rdtypes.dnskeybase import Flag

import dns.dnssec
import dns.name
import dns.rdataclass
import dns.rdatatype
import dns.rrset
import dns.zonefile

from .algorithms import (
    ALL_ALGORITHMS_BY_NUM,
    ECDSAP256SHA256,
    ECDSAP384SHA384,
    ED448,
    ED25519,
    Algorithm,
)
from .log import debug
from .run import EnvCmd
from .template import NS1, Nameserver, TemplateEngine, TrustAnchor

KEYDIR = "keys"
DNSKEY_TTL = 3600

PrivateKey: TypeAlias = (
    ec.EllipticCurvePrivateKey
    | ed25519.Ed25519PrivateKey
    | ed448.Ed448PrivateKey
    | rsa.RSAPrivateKey
)


class ZoneKey(ABC):
    """
    Abstract base for a DNSSEC key attached to a Zone.

    Two concrete implementations exist:
      FileZoneKey    — reads a dnssec-keygen-managed key file pair.
      PythonZoneKey  — holds a Python-native (private_key, dnskey_rdata) pair
                       required for dnspython-based operations and signing.

    The interface covers the zone-infrastructure subset: trust anchor
    generation and DS propagation to parent zones.
    """

    @property
    @abstractmethod
    def dnskey(self) -> dns.rrset.RRset:
        """
        The DNSKEY RRset for this key (single-record, with TTL).
        """

    @property
    @abstractmethod
    def private_key(self) -> PrivateKey:
        """
        The dnspython private-key object, for use with dns.dnssec signing.
        """

    def is_ksk(self) -> bool:
        return bool(self.dnskey[0].flags & int(Flag.SEP))

    @abstractmethod
    def write_dsset(
        self,
        target_dir: Path | str,
        dsdigest: dns.dnssec.DSDigest = dns.dnssec.DSDigest.SHA256,
    ) -> None:
        """
        Write or copy dsset-{zone}. into target_dir.

        For FileZoneKey: copies the dsset file produced by dnssec-signzone.
        For PythonZoneKey: derives the DS from the in-memory key and writes it.
        """

    def into_ta(
        self,
        ta_type: str = "static-ds",
        dsdigest: dns.dnssec.DSDigest = dns.dnssec.DSDigest.SHA256,
    ) -> TrustAnchor:
        """
        Build a named.conf trust-anchor stanza from this key.

        ta_type must be one of: static-ds, initial-ds, static-key, initial-key.
        Implemented once here; both subclasses inherit it via self.dnskey.
        """
        dnskey = self.dnskey
        if ta_type in ["static-ds", "initial-ds"]:
            ds = dns.dnssec.make_ds(dnskey.name, dnskey[0], dsdigest)
            parts = str(ds).split()
            contents = " ".join(parts[:3]) + f' "{parts[3]}"'
        elif ta_type in ["static-key", "initial-key"]:
            parts = str(dnskey).split()
            contents = " ".join(parts[4:7]) + f' "{"".join(parts[7:])}"'
        else:
            raise ValueError(f"invalid trust anchor type: {ta_type!r}")
        return TrustAnchor(str(dnskey.name), ta_type, contents)


class FileZoneKey(ZoneKey):
    """
    A ZoneKey backed by dnssec-keygen-managed key files.

    Reads key material directly from the .key/.private file pair.  Normally
    constructed via FileZoneKey.generate() (or Zone.add_keys());
    isctest.kasp.Key subclasses it to add KASP timing/state operations.
    """

    def __init__(
        self,
        name: str,
        keydir: str | Path | None = None,
        zone: Zone | None = None,
    ) -> None:
        self.name = name
        self.keydir = Path() if keydir is None else Path(keydir)
        self.path = str(self.keydir / name)
        self.privatefile = f"{self.path}.private"
        self.keyfile = f"{self.path}.key"
        self.tag = int(self.name[-5:])
        self.zone = zone

    @property
    def dnskey(self) -> dns.rrset.RRset:
        with open(self.keyfile, "r", encoding="utf-8") as file:
            rrsets = dns.zonefile.read_rrsets(
                file.read(),
                rdclass=None,  # read rdclass from the file
                default_ttl=DNSKEY_TTL,  # use this TTL if not present
            )
        assert len(rrsets) == 1, f"{self.keyfile} has multiple RRsets"
        dnskey_rr = rrsets[0]
        assert len(dnskey_rr) == 1, f"{self.keyfile} has multiple RRs"
        assert (
            dnskey_rr.rdtype == dns.rdatatype.DNSKEY
        ), f"DNSKEY not found in {self.keyfile}"
        return dnskey_rr

    @property
    def private_key(self) -> PrivateKey:
        """
        Read the .private file and build the corresponding cryptography
        private-key object, dispatching on the DNSKEY algorithm.

        Supports the RSA, ECDSA and EdDSA families; other algorithms raise
        NotImplementedError.
        """
        alg = ALL_ALGORITHMS_BY_NUM[self.dnskey[0].algorithm]

        fields = {}
        with open(self.privatefile, "r", encoding="utf-8") as file:
            for line in file:
                name, sep, value = line.partition(":")
                if sep:
                    fields[name.strip()] = value.strip()

        def field(name: str) -> bytes:
            return base64.b64decode(fields[name])

        if "Modulus" in fields:  # RSA family, including the private-OID variants
            public = rsa.RSAPublicNumbers(
                int.from_bytes(field("PublicExponent"), "big"),
                int.from_bytes(field("Modulus"), "big"),
            )
            return rsa.RSAPrivateNumbers(
                p=int.from_bytes(field("Prime1"), "big"),
                q=int.from_bytes(field("Prime2"), "big"),
                d=int.from_bytes(field("PrivateExponent"), "big"),
                dmp1=int.from_bytes(field("Exponent1"), "big"),
                dmq1=int.from_bytes(field("Exponent2"), "big"),
                iqmp=int.from_bytes(field("Coefficient"), "big"),
                public_numbers=public,
            ).private_key()

        secret = field("PrivateKey")
        if alg == ECDSAP256SHA256:
            return ec.derive_private_key(int.from_bytes(secret, "big"), ec.SECP256R1())
        if alg == ECDSAP384SHA384:
            return ec.derive_private_key(int.from_bytes(secret, "big"), ec.SECP384R1())
        if alg == ED25519:
            return ed25519.Ed25519PrivateKey.from_private_bytes(secret)
        if alg == ED448:
            return ed448.Ed448PrivateKey.from_private_bytes(secret)
        raise NotImplementedError(f"dnskey algorithm {alg.name} not implemented")

    def write_dsset(
        self,
        target_dir: Path | str,
        dsdigest: dns.dnssec.DSDigest = dns.dnssec.DSDigest.SHA256,
    ) -> None:
        """
        Copy the dnssec-signzone-produced dsset-{zone}. into target_dir.

        dsdigest is accepted for interface compatibility but ignored: the dsset
        file is produced by dnssec-signzone (SHA-256). This copy overwrites any
        existing dsset file, so a zone must not mix FileZoneKey and
        PythonZoneKey KSKs (PythonZoneKey appends to the same file);
        Zone.copy_dssets enforces this.
        """
        assert self.zone is not None, "write_dsset requires a zone-attached key"
        src = Path(self.zone.ns.name) / f"dsset-{self.zone.name}."
        dst = Path(target_dir) / src.name
        if src.resolve() == dst.resolve():
            debug(f"{self.zone.name}: dsset already in {target_dir}")
            return
        shutil.copy(src, dst)
        debug(f"{self.zone.name}: dsset copied to {target_dir}")

    @staticmethod
    def generate(
        zone: Zone,
        params: str = "",
        alg: Algorithm | None = None,
    ) -> FileZoneKey:
        """
        Generate a DNSSEC key via dnssec-keygen for zone and return it.

        Runs dnssec-keygen in zone.ns.name/keys/, stores the key there, and
        returns the resulting FileZoneKey.  Pass params="-f KSK" to generate a
        Key Signing Key; omit it (or pass "") for a Zone Signing Key.
        """
        debug(f"{zone.name}: generating key using dnssec-keygen")
        keydir = Path(zone.ns.name) / KEYDIR
        keydir.mkdir(exist_ok=True)
        if alg is None:
            alg = Algorithm.default()
        keygen = EnvCmd(
            "KEYGEN", f"-q -a {alg.number} -b {alg.bits} -K {KEYDIR} -L {DNSKEY_TTL}"
        )
        key_name = keygen(f"{params} {zone.name}", cwd=zone.ns.name).out.strip()
        return FileZoneKey(key_name, keydir=keydir, zone=zone)


class PythonZoneKey(ZoneKey):
    """
    A ZoneKey holding a Python-native keypair.

    Construct via PythonZoneKey.generate() to create fresh key
    material, or instantiate directly when you already have a private key and
    dnskey rdata (e.g. when loading a saved PEM).

    Attach to a Zone via zone.keys = [key] so that Zone.copy_dssets() can
    generate the dsset-* file for the parent zone and Zone.trust_anchors() can
    produce the correct trust anchor stanzas.

    Zone.sign() raises TypeError if self.keys contains a PythonZoneKey,
    because dnssec-signzone cannot use in-memory keys.  Sign the zone
    with dns.dnssec.sign_zone() directly instead.

    The raw private key object is available as .private_key for callers that
    need it (e.g. to write a PEM file for a custom authoritative server).
    Use write_private_key_pem() as a convenience for the common case.
    """

    def __init__(
        self,
        zone: Zone,
        private_key,
        dnskey_rdata,
        ttl: int = DNSKEY_TTL,
    ) -> None:
        self.zone = zone
        self._private_key = private_key
        self._dnskey_rdata = dnskey_rdata
        self.ttl = ttl

    @property
    def private_key(self) -> PrivateKey:
        return self._private_key

    @property
    def dnskey(self) -> dns.rrset.RRset:
        rrset = dns.rrset.RRset(
            self.zone.dname, dns.rdataclass.IN, dns.rdatatype.DNSKEY
        )
        rrset.update_ttl(self.ttl)
        rrset.add(self._dnskey_rdata)
        return rrset

    def write_dsset(
        self,
        target_dir: Path | str,
        dsdigest: dns.dnssec.DSDigest = dns.dnssec.DSDigest.SHA256,
    ) -> None:
        target = Path(target_dir)
        target.mkdir(parents=True, exist_ok=True)
        ds = dns.dnssec.make_ds(self.zone.dname, self._dnskey_rdata, dsdigest)
        text = (
            f"{self.zone.name}. {self.ttl} IN DS"
            f" {ds.key_tag} {ds.algorithm} {ds.digest_type}"
            f" {ds.digest.hex().upper()}\n"
        )
        with (target / f"dsset-{self.zone.name}.").open("a") as f:
            f.write(text)

    def write_private_key_pem(self, path: Path | str) -> None:
        """Write the private key to path in PKCS8 PEM format (no encryption)."""
        Path(path).write_bytes(
            self.private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )

    @staticmethod
    def generate(
        zone: Zone,
        flags: int = int(Flag.ZONE | Flag.SEP),
        alg: Algorithm | None = None,
        ttl: int = DNSKEY_TTL,
    ) -> PythonZoneKey:
        """
        Generate a Python-native DNSSEC keypair for the given algorithm.

        Unlike FileZoneKey.generate(), this does not invoke
        dnssec-keygen and produces no on-disk key files.  The returned
        PythonZoneKey is suitable for use with dns.dnssec.sign(),
        dns.dnssec.sign_zone(), and dns.dnssec.make_ds().

        The algorithm-to-key-type mapping is:
          ECDSAP256SHA256 -> EC P-256
          ECDSAP384SHA384 -> EC P-384
          ED25519         -> Ed25519
          ED448           -> Ed448
          RSASHA1/256/512 -> RSA (key size from alg.bits)

        Args:
            zone:  The Zone to generate the keypair for.
            flags: DNSKEY flags bitmask; defaults to ZONE|SEP (KSK).
            alg:   Algorithm to use; defaults to Algorithm.default().
            ttl:   TTL for the DNSKEY RRset (default DNSKEY_TTL).
        """
        if alg is None:
            alg = Algorithm.default()
        _generators: dict[str, Callable[[], PrivateKey]] = {
            "ECDSAP256SHA256": lambda: ec.generate_private_key(ec.SECP256R1()),
            "ECDSAP384SHA384": lambda: ec.generate_private_key(ec.SECP384R1()),
            "ED25519": ed25519.Ed25519PrivateKey.generate,
            "ED448": ed448.Ed448PrivateKey.generate,
            "RSASHA1": lambda: rsa.generate_private_key(65537, alg.bits),
            "RSASHA256": lambda: rsa.generate_private_key(65537, alg.bits),
            "RSASHA512": lambda: rsa.generate_private_key(65537, alg.bits),
        }
        gen = _generators.get(alg.name)
        if gen is None:
            raise ValueError(
                f"unsupported algorithm for Python-native key generation: {alg.name!r}"
            )
        private_key = gen()
        dnskey_rdata = dns.dnssec.make_dnskey(
            private_key.public_key(),
            dns.dnssec.Algorithm(alg.number),
            flags=flags,
        )
        return PythonZoneKey(zone, private_key, dnskey_rdata, ttl)


class Zone:
    """
    Zone providing zone file setup and signing operations.

    This is the operational counterpart to isctest.template.Zone, which is a
    plain data container for template rendering. Use isctest.template.zones()
    to convert a list of Zone instances into template data.

    The normal entrypoint is configure(), which runs the full setup once.
    The individual steps (copy_dssets, add_keys, render, sign) are public so
    that tests needing finer-grained control can drive them directly, but they
    are order-dependent and write to the same on-disk locations as configure().
    Pick one or the other for a given Zone: calling a step method and
    configure() on the same Zone leaves the zone directory in an inconsistent
    state.
    """

    def __init__(
        self,
        name: str | dns.name.Name,
        ns: Nameserver,
        signed: bool = False,
        subdir: str | None = "zones",
        filepath_unsigned: Path | str | None = None,
        filepath_signed: Path | str | None = None,
        zone_type: str = "primary",
    ) -> None:
        self.dname: dns.name.Name = (
            dns.name.from_text(name) if isinstance(name, str) else name
        )
        raw = self.dname.to_text()
        self.name: str = raw if raw == "." else raw.rstrip(".")
        self.basename: str = "root" if self.dname == dns.name.root else self.name
        self.ns = ns
        self.signed = signed
        self.subdir = subdir
        self.type = zone_type
        self._configured = False

        prefix = f"{subdir}/" if subdir else ""
        self.filepath_unsigned: Path = Path(
            filepath_unsigned or f"{prefix}{self.basename}.db"
        )
        self.filepath_signed: Path = Path(
            filepath_signed or f"{prefix}{self.basename}.db.signed"
        )

        self.delegations: list[Zone] = []
        self.keys: list[ZoneKey] = []

    @property
    def filepath(self) -> Path:
        """Actual zone file — filepath_signed if signed, filepath_unsigned otherwise."""
        return self.filepath_signed if self.signed else self.filepath_unsigned

    def add_keys(self, ksk: bool = True, zsk: bool = True) -> None:
        """Generate KSK and/or ZSK via dnssec-keygen and append to self.keys."""
        if ksk:
            self.keys.append(FileZoneKey.generate(self, "-f KSK"))
        if zsk:
            self.keys.append(FileZoneKey.generate(self))

    def copy_dssets(self) -> None:
        """Write dsset-* files for each signed delegation into self.ns dir."""
        for zone in self.delegations:
            ksks = [k for k in zone.keys if k.is_ksk()]
            has_file = any(isinstance(k, FileZoneKey) for k in ksks)
            has_python = any(isinstance(k, PythonZoneKey) for k in ksks)
            if has_file and has_python:
                raise TypeError(
                    f"{zone.name}: cannot mix FileZoneKey and PythonZoneKey KSKs; "
                    "dsset writing is order-dependent (FileZoneKey overwrites, "
                    "PythonZoneKey appends)"
                )
            if ksks:
                for key in ksks:
                    key.write_dsset(Path(self.ns.name))
            else:
                debug(f"{zone.name}: delegation is insecure (no KSK)")

    def render(self, template: str | None = None) -> None:
        """Render the unsigned zone file from a jinja2 template."""
        debug(f"{self.name}: rendering zone file")
        templates = TemplateEngine(".")
        output = Path(self.ns.name) / self.filepath_unsigned
        output.parent.mkdir(exist_ok=True)

        if template is None:
            local = f"{output}.j2.manual"
            common = "_common/zones/template.db.j2.manual"
            template = local if Path(local).is_file() else common

        data = {
            "zone": self,
            "delegations": self.delegations,
        }
        templates.render(str(output), data, template=template)

    def sign(self, params: str = "") -> None:
        """
        Sign the rendered zone file via dnssec-signzone.

        Requires self.signed == True.  Raises TypeError if self.keys contains
        any PythonZoneKey — dnssec-signzone cannot use in-memory keys; use
        dns.dnssec.sign_zone() directly for Python-native signing.
        """
        assert self.signed, f"{self.name}: zone is not configured for signing"
        python_keys = [k for k in self.keys if isinstance(k, PythonZoneKey)]
        if python_keys:
            raise TypeError(
                f"{self.name}: Zone.sign() invokes dnssec-signzone which requires "
                "file-backed keys; use dns.dnssec.sign_zone() for Python-native keys"
            )
        debug(f"{self.name}: signing zone")
        signer = EnvCmd("SIGNER", f"-S -g -K {KEYDIR} {params}")
        signer(
            f"-P -x -O full -o {self.name}"
            f" -f {self.filepath_signed} {self.filepath_unsigned}",
            cwd=self.ns.name,
        )

    def configure(self, template: str | None = None, sign_params: str = "") -> None:
        """
        Perform full zone setup: copy DS sets, generate keys, render, sign.

        This is the standard single-call entrypoint and may be called only once
        per Zone. Use the individual step methods directly only when a test
        needs finer-grained control, and do not mix them with configure() on the
        same Zone (see the class docstring).
        """
        assert not self._configured, f"{self.name}: configure() already called"
        self._configured = True
        self.copy_dssets()
        if self.signed:
            self.add_keys()
        self.render(template)
        if self.signed:
            self.sign(sign_params)

    def trust_anchors(self, ta_type: str = "static-ds") -> list[TrustAnchor]:
        """Return a trust-anchor stanza for every KSK in zone.keys."""
        ksks = [k for k in self.keys if k.is_ksk()]
        assert ksks, f"{self.name}: no KSK in zone.keys"
        return [k.into_ta(ta_type) for k in ksks]


def configure_root(
    delegations: list[Zone],
    ns: Nameserver = NS1,
    signed: bool = True,
) -> Zone:
    zone = Zone(".", ns, signed=signed)
    zone.delegations = delegations
    zone.configure(template="_common/zones/root.db.j2.manual")
    return zone



"""
Copyright (C) Internet Systems Consortium, Inc. ("ISC")

SPDX-License-Identifier: MPL-2.0

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0.  If a copy of the MPL was not distributed with this
file, you can obtain one at https://mozilla.org/MPL/2.0/.

See the COPYRIGHT file distributed with this work for additional
information regarding copyright ownership.
"""

from collections.abc import AsyncGenerator

import abc

import dns.flags
import dns.message
import dns.rcode
import dns.rdatatype

from isctest.asyncserver import (
    AsyncDnsServer,
    DnsProtocol,
    DnsResponseSend,
    DomainHandler,
    QnameHandler,
    QnameQtypeHandler,
    QueryContext,
    ResponseHandler,
    StaticResponseHandler,
)

from ..resolver_ans import rrset


class HeaderOnlyHandler(ResponseHandler):
    """
    Return an empty DNS message with only header flags set.
    """

    @property
    @abc.abstractmethod
    def flags(self) -> dns.flags.Flag:
        raise NotImplementedError

    @property
    def rcode(self) -> dns.rcode.Rcode:
        return dns.rcode.NOERROR

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[DnsResponseSend, None]:
        message = dns.message.Message(id=qctx.query.id)
        message.use_edns(False)
        message.flags = self.flags
        message.set_rcode(self.rcode)
        yield DnsResponseSend(message, acknowledge_hand_rolled_response=True)


class RefusedOnTcpHandler(QnameHandler, HeaderOnlyHandler):
    qnames = ["tcpalso.no-questions."]
    flags = dns.flags.QR
    rcode = dns.rcode.REFUSED

    def match(self, qctx: QueryContext) -> bool:
        return qctx.protocol == DnsProtocol.TCP and super().match(qctx)


class TcpFallbackHandler(ResponseHandler):
    def match(self, qctx: QueryContext) -> bool:
        return qctx.protocol == DnsProtocol.TCP

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[DnsResponseSend, None]:
        qctx.response.answer.append(rrset(qctx.qname, dns.rdatatype.A, "1.2.3.4"))
        yield DnsResponseSend(qctx.response)


class FormerrToAllHandler(DomainHandler, StaticResponseHandler):
    domains = ["formerr-to-all."]
    rcode = dns.rcode.FORMERR


class NoQuestionsNSHandler(QnameQtypeHandler, StaticResponseHandler):
    qnames = ["no-questions."]
    qtypes = [dns.rdatatype.NS]
    answer = [rrset(qnames[0], dns.rdatatype.NS, f"ns.{qnames[0]}")]
    additional = [rrset(f"ns.{qnames[0]}", dns.rdatatype.A, "10.53.0.8")]


class NsNoQuestionsAHandler(QnameHandler):
    qnames = ["ns.no-questions."]

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[DnsResponseSend, None]:
        if qctx.qtype == dns.rdatatype.A:
            a_rrset = rrset(qctx.qname, dns.rdatatype.A, "10.53.0.8")
            qctx.response.answer.append(a_rrset)
        yield DnsResponseSend(qctx.response)


class TcpalsoNoQuestionsHandler(QnameHandler, HeaderOnlyHandler):
    qnames = ["tcpalso.no-questions."]
    flags = dns.flags.QR | dns.flags.TC
    rcode = dns.rcode.REFUSED


class TruncatedNoQuestionsHandler(QnameHandler, HeaderOnlyHandler):
    qnames = ["truncated.no-questions."]
    flags = dns.flags.QR | dns.flags.AA | dns.flags.TC


class FallbackHandler(HeaderOnlyHandler):
    flags = dns.flags.QR | dns.flags.AA


def main() -> None:
    server = AsyncDnsServer(default_aa=True, default_rcode=dns.rcode.NOERROR)

    # Install TCP handlers first so they take precedence
    server.install_response_handlers(
        RefusedOnTcpHandler(),
        TcpFallbackHandler(),
    )

    # Install UDP handlers
    server.install_response_handlers(
        FormerrToAllHandler(),
        NoQuestionsNSHandler(),
        NsNoQuestionsAHandler(),
        TcpalsoNoQuestionsHandler(),
        TruncatedNoQuestionsHandler(),
    )
    server.install_response_handler(FallbackHandler())

    server.run()


if __name__ == "__main__":
    main()



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from re import compile as Re

from dnssec_py.common import DNSSEC_PY_MARK
from isctest.template import NS2, TrustAnchor, zones
from isctest.zone import Zone, configure_root

import isctest

pytestmark = DNSSEC_PY_MARK


def bootstrap():
    revoked_zone = Zone("truncated-revoked.selfsigned", NS2, signed=True)
    active_zone = Zone("truncated-active.selfsigned", NS2, signed=True)

    # just delegation, the trust anchors are added directly
    root = configure_root([revoked_zone, active_zone], signed=False)

    # The trust anchor key tag must match the revoked truncated self-signed key
    # in the zone (key tag 33167). The flags differ here (257 vs 385) because
    # the revoked bit is not part of the trust anchor, but it is part of the key
    # tag calculation.
    revoked_ta = TrustAnchor(
        "truncated-revoked.selfsigned", "static-key", '257 3 14 "fYA="'
    )

    # The active truncated key is too short for the ECDSA curve but passes the
    # parser's minimum-length check; trusting it directly exercises the
    # key-construction failure path.
    active_ta = TrustAnchor(
        "truncated-active.selfsigned", "static-key", '257 3 14 "fYA="'
    )

    return {
        "trust_anchors": [revoked_ta, active_ta],
        "zones": zones([root, revoked_zone, active_zone]),
    }


def test_truncated_revoked_dnskey(ns9):
    msg = isctest.query.create("a.truncated-revoked.selfsigned.", "A")
    with ns9.watch_log_from_here() as watcher:
        res = isctest.query.tcp(msg, ns9.ip)
        watcher.wait_for_line(
            Re("a.truncated-revoked.selfsigned/A.*broken trust chain")
        )
    isctest.check.servfail(res)


def test_truncated_active_dnskey(ns9):
    msg = isctest.query.create("a.truncated-active.selfsigned.", "A")
    res = isctest.query.tcp(msg, ns9.ip)
    isctest.check.servfail(res)



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

"""
Regression test for GSS-API context leak via repeated TKEY queries.

An unauthenticated attacker could exhaust server memory by sending
repeated TKEY queries with crafted SPNEGO NegTokenInit tokens.
Each query triggers gss_accept_sec_context() which returns
GSS_S_CONTINUE_NEEDED and allocates a GSS context.  On the unfixed
code path, the context handle in process_gsstkey() is never stored
or freed, leaking ~520 bytes per query.

The fix rejects GSS_S_CONTINUE_NEEDED in dst_gssapi_acceptctx() and
deletes the context immediately.

The key distinguishing signal in the TKEY response:
  - CONTINUE (vulnerable): error=0, output token present, no TSIG
  - BADKEY (fixed):        error=17, no output token
"""

import struct
import time

import dns.name
import dns.query
import dns.rdataclass
import dns.rdatatype
import dns.rdtypes.ANY.TKEY
import pytest

import isctest

pytestmark = pytest.mark.extra_artifacts(
    [
        "*/*.db",
    ]
)

TKEY_NAME = dns.name.from_text("test.key.")
GSSAPI_ALGORITHM = dns.name.from_text("gss-tsig.")
TKEY_MODE_GSSAPI = 3

# OID 1.2.840.113554.1.2.2 (Kerberos 5)
KRB5_OID = b"\x06\x09\x2a\x86\x48\x86\xf7\x12\x01\x02\x02"

# OID 1.3.6.1.5.5.2 (SPNEGO)
SPNEGO_OID = b"\x06\x06\x2b\x06\x01\x05\x05\x02"


def der_encode(tag, data):
    """Encode data in ASN.1 DER TLV format."""
    length = len(data)
    if length < 128:
        return tag + bytes([length]) + data
    if length < 256:
        return tag + b"\x81" + bytes([length]) + data
    return tag + b"\x82" + struct.pack(">H", length) + data


def spnego_negtokeninit():
    """Build a SPNEGO NegTokenInit proposing krb5 without a mechToken.

    This forces gss_accept_sec_context() to return GSS_S_CONTINUE_NEEDED
    because the acceptor recognizes the krb5 mechanism but has not
    received an actual AP-REQ token yet.
    """
    # MechTypeList ::= SEQUENCE OF MechType
    mechtype_list = der_encode(b"\x30", KRB5_OID)
    # [0] mechTypes
    mechtypes = der_encode(b"\xa0", mechtype_list)
    # NegTokenInit ::= SEQUENCE { mechTypes, ... }
    negtokeninit = der_encode(b"\x30", mechtypes)
    # [0] CONSTRUCTED (wrapping NegTokenInit)
    wrapped = der_encode(b"\xa0", negtokeninit)
    # APPLICATION 0 CONSTRUCTED (SPNEGO OID + body)
    return der_encode(b"\x60", SPNEGO_OID + wrapped)


def make_tkey_query(token):
    """Build a TKEY query with a GSS-API token in the additional section."""
    now = int(time.time())
    tkey_rdata = dns.rdtypes.ANY.TKEY.TKEY(
        rdclass=dns.rdataclass.ANY,
        rdtype=dns.rdatatype.TKEY,
        algorithm=GSSAPI_ALGORITHM,
        inception=now,
        expiration=now + 86400,
        mode=TKEY_MODE_GSSAPI,
        error=0,
        key=token,
        other=b"",
    )

    msg = isctest.query.create(TKEY_NAME, dns.rdatatype.TKEY, dns.rdataclass.ANY)
    rrset = msg.find_rrset(
        msg.additional,
        TKEY_NAME,
        dns.rdataclass.ANY,
        dns.rdatatype.TKEY,
        create=True,
    )
    rrset.add(tkey_rdata)
    return msg


def test_tkey_gssapi_no_continuation(ns1):
    """TKEY with a SPNEGO NegTokenInit must be rejected, not continued.

    On unfixed code, gss_accept_sec_context() returns CONTINUE_NEEDED
    and the response has error=0 with an output token (the leaked path).
    On fixed code, CONTINUE_NEEDED is rejected and the response has
    error=BADKEY(17) with no output token.
    """
    port = ns1.ports.dns
    ip = ns1.ip

    msg = make_tkey_query(spnego_negtokeninit())
    res = dns.query.tcp(msg, ip, port=port, timeout=5)

    assert res is not None

    tkey = get_tkey_answer(res)
    assert tkey is not None, "server did not return a TKEY answer"
    assert (
        tkey.error != 0
    ), "server returned error=0 (GSS_S_CONTINUE_NEEDED not rejected)"
    assert len(tkey.key) == 0, "server returned a continuation token"


def get_tkey_answer(response):
    """Extract TKEY rdata from a DNS response, or None."""
    for rrset in response.answer:
        if rrset.rdtype == dns.rdatatype.TKEY:
            for rdata in rrset:
                return rdata
    return None



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from collections.abc import AsyncGenerator

import logging

import dns.rcode
import dns.rdatatype
import dns.rrset

from isctest.asyncserver import (
    CloseConnection,
    ControlCommand,
    ControllableAsyncDnsServer,
    DnsResponseSend,
    QueryContext,
    ResponseAction,
    ResponseHandler,
)


class ErraticAxfrHandler(ResponseHandler):
    allowed_actions = ["no-response", "partial-axfr", "complete-axfr"]

    def __init__(self, actions: list[str]) -> None:
        self.actions = actions
        self.counter = 0
        for action in actions:
            assert action in self.allowed_actions

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[ResponseAction, None]:
        action = self.actions[self.counter % len(self.actions)]
        self.counter += 1

        logging.info("current response action: %s", action)

        if action == "no-response":
            yield CloseConnection()
            return

        soa_rr = dns.rrset.from_text(
            qctx.qname, 300, qctx.qclass, dns.rdatatype.SOA, ". . 0 0 0 0 300"
        )
        ns_rr = dns.rrset.from_text(qctx.qname, 300, qctx.qclass, dns.rdatatype.NS, ".")

        qctx.response.answer.append(soa_rr)
        qctx.response.answer.append(ns_rr)

        if action == "partial-axfr":
            yield DnsResponseSend(qctx.response)
        elif action == "complete-axfr":
            qctx.response.answer.append(soa_rr)
            yield DnsResponseSend(qctx.response)
        yield CloseConnection()


class ResponseSequenceCommand(ControlCommand):
    control_subdomain = "response-sequence"

    def __init__(self) -> None:
        self._current_handler: ResponseHandler | None = None

    def handle(
        self, args: list[str], server: ControllableAsyncDnsServer, qctx: QueryContext
    ) -> str:
        for action in args:
            if action not in ErraticAxfrHandler.allowed_actions:
                logging.error("invalid %s action '%s'", self, action)
                qctx.response.set_rcode(dns.rcode.SERVFAIL)
                return f"invalid action '{action}'; must be one of {ErraticAxfrHandler.allowed_actions}"

        actions = args

        if self._current_handler is not None:
            server.uninstall_response_handler(self._current_handler)

        self._current_handler = ErraticAxfrHandler(actions)
        server.install_response_handler(self._current_handler)

        msg = f"reponse sequence set to {actions}"
        logging.info(msg)
        return msg


def main() -> None:
    server = ControllableAsyncDnsServer(
        default_aa=True, default_rcode=dns.rcode.NOERROR
    )
    server.install_control_command(ResponseSequenceCommand())
    server.run()


if __name__ == "__main__":
    main()
